import os
import logging
import sys
import traceback
from typing import List, Optional
import click
import signal
import time
from functools import partial

from logging.handlers import RotatingFileHandler
from dataspree.inference.core.exceptions import StopTimeout

from dataspree.inference import __version__


def sleep(seconds: float):
    time_wake_up: float = time.time() + seconds
    try:
        while True:
            time_to_sleep = min(max(time_wake_up - time.time(), 0.0), 0.5)
            if time_to_sleep > 0:
                time.__dataspree_inference_sleep(time_to_sleep)
            else:
                return
    except StopTimeout:
        raise StopTimeout('Node stopped.')


time.__dataspree_inference_sleep = time.sleep
time.sleep = sleep


# noinspection PyPep8Naming
class SM(staticmethod):
    """
    Quick workaround that allows us to use Cython and multiple decorators.
    """

    def __init__(self, *args, **kwargs):
        super(staticmethod, self).__init__(*args, **kwargs)
        self.__ds_in_static_hack__ = args[0]


if hasattr(globals()['__builtins__'], 'staticmethod'):
    globals()['__builtins__'].staticmethod = SM


def terminate_application(application, *_args) -> None:
    application.running = False


@click.group()
@click.version_option(__version__)
def cli():
    pass


@cli.command('visualize_log')
@click.option('--input-files', '-i', multiple=True, required=True, help='Input log files',
              type=click.Path(exists=True, dir_okay=False, file_okay=True))
@click.option('--plot_cpu', '-c', is_flag=True, required=False, type=bool,
              help='Plot CPU utilization statistics')
@click.option('--plot_memory', '-m', is_flag=True, required=False, type=bool,
              help='Plot memory utilization statistics')
@click.option('--plot_network', '-n', is_flag=True, required=False, type=bool,
              help='Plot network I/O')
@click.option('--plot_disk_io', '-d', is_flag=True, required=False, type=bool,
              help='Plot process\' disk I/O ')
@click.option('--plot_frame_drop', '-fd', is_flag=True, required=False, type=bool,
              help='Plot node\'s frame drop statistics')
@click.option('--plot_frame_rate', '-fr', is_flag=True, required=False, type=bool,
              help='Plot frame rate at node.')
@click.option('--plot_frame_io_time', '-fio', is_flag=True, required=False, type=bool,
              help='Plot frame I/O time at a node.')
@click.option('--plot_cumulative_node_latency', '-cl', is_flag=True, required=False, type=bool,
              help='Plot the time a frame took from its creation to the output of the specific '
                   'plugins.')
@click.option('--plot_node_latency', '-l', is_flag=True, required=False, type=bool, help='Plot the frames\' '
                                                                                         'processing time at '
                                                                                         'specific nodes')
@click.option('--save', '-s', is_flag=True, required=False, type=bool)
def visualize_log(input_files: List[str],
                  plot_cpu: bool = False,
                  plot_memory: bool = False,
                  plot_network: bool = False,
                  plot_disk_io: bool = False,
                  plot_frame_drop: bool = False,
                  plot_frame_rate: bool = False,
                  plot_frame_io_time: bool = False,
                  plot_cumulative_node_latency: bool = True,
                  plot_node_latency: bool = True,
                  save: bool = False):
    from dataspree.inference.core.profiler.csv_output import CsvOutput
    CsvOutput.visualize(input_files, plot_cpu=plot_cpu, plot_memory=plot_memory,
                        plot_network=plot_network, plot_disk_io=plot_disk_io,
                        plot_frame_drop=plot_frame_drop, plot_frame_rate=plot_frame_rate,
                        plot_frame_io_time=plot_frame_io_time,
                        plot_cumulative_node_latency=plot_cumulative_node_latency,
                        plot_node_latency=plot_node_latency, save=save)


@cli.command('simplify')
@click.option('--input_file', type=str, help='Path to the input model onnx file', multiple=False, required=True)
@click.option('--output_file', type=str, help='Path to the desired output onnx file', multiple=False, default=None)
def simplify(input_file: Optional[str], output_file: Optional[str]) -> None:
    try:
        from onnxsim import simplify
    except ImportError as e:
        print(f'Could not import required module: {e}. Please install it.')
        sys.exit(-1)

    import onnx

    # Override input in case the simplification succeeds.
    if output_file is None:
        output_file = input_file

    model = onnx.load(input_file)
    model_simp, check = simplify(model, dynamic_input_shape=True)

    if not check:
        raise 'The model could not be exported'

    onnx.save_model(model_simp, output_file)
    print(f'Model simplified and written to {output_file}')


@cli.command('run')
@click.option('--debug', is_flag=True, help='Enable debugging mode. This influences error handling '
                                            'and the default log level setting.')
@click.option('--log_level', type=click.Choice(['debug', 'info', 'warning', 'error']), required=False, default=None)
@click.option('--organization', type=str,
              help='Specify the organization to which this Dataspree Inference instance is licensed.',
              multiple=False, default=None)
def run(debug: bool, log_level: Optional[str] = None, organization: Optional[str] = None) -> None:
    print(
        r''' __       ___          __   __   __   ___  ___
|  \  /\   |   /\     /__` |__) |__) |__  |__ 
|__/ /--\  |  /--\    .__/ |    |  \ |___ |___

Inference
    ''')
    log_format = '%(asctime)s %(name)-52s %(levelname)-6s %(message)s'

    if log_level is not None:
        logger_level: int = {
            'debug': logging.DEBUG,
            'info': logging.INFO,
            'warning': logging.WARNING,
            'error': logging.ERROR,
        }[log_level]

        logging.basicConfig(level=logger_level, format=log_format)
    else:
        if debug:
            logging.basicConfig(level=logging.DEBUG, format=log_format)
        else:
            logging.basicConfig(level=logging.INFO, format=log_format)

    logging.getLogger('aiohttp.access').setLevel(logging.WARNING)
    logging.getLogger('asyncua').setLevel(logging.WARNING)
    logging.getLogger('numba').setLevel(logging.WARNING)

    application = None

    def new_except_hook(error_type, error_value, error_traceback) -> None:
        """
        Exception hook which logs uncaught exceptions to the logger output.

        @param error_type:  Type of the exception that is thrown but uncaught
        @param error_value: Instance of the exception that is thrown but uncaught, which contains
                            a specification on what actually happened.
        @param error_traceback:  Traceback, which we use to print debug information when available.
        """

        if debug:
            sys.__excepthook__(error_type, error_value, error_traceback)

        format_traceback = '\n'.join([str(k) for k in traceback.walk_tb(error_traceback)])
        logging.getLogger().fatal(f'An unexpected error occurred:\n   Type: {error_type}\n  '
                                  f'Value: {error_value}\n   Traceback: \n{format_traceback}')

    sys.excepthook = new_except_hook

    try:

        from dataspree.inference.core.utils import log_dir
        log_file = os.path.join(log_dir(), 'application.log')
        rotating_file_handler = RotatingFileHandler(log_file, maxBytes=25 * 1024 * 1024,
                                                    backupCount=10)
        log_formatter = logging.Formatter('%(asctime)s %(name)s %(levelname)s %(message)s')
        rotating_file_handler.setFormatter(log_formatter)
        logging.getLogger().addHandler(rotating_file_handler)

        from dataspree.inference.core.application import Application
        application = Application(debug=debug, organization=organization)
        signal.signal(signal.SIGINT, partial(terminate_application, application))
        application.run()

    except KeyboardInterrupt:
        logging.info('Keyboard interrupt')
    except Exception as e:
        logging.error(f'An exception occurred: {e}')
        if debug:
            raise e

    if application is not None:
        logging.info('Stopping application...')
        application.stop()
    logging.info('Shutdown successful.')


if __name__ == '__main__':
    cli()
