import time
from typing import Dict, Optional, List, Tuple

import cv2
import numpy as np
from PIL import Image, ImageDraw
from dataspree.inference.core.config import ConfigFieldTypes
from dataspree.inference.core.consumer import Consumer
from dataspree.inference.core.item import Item
from dataspree.inference.core.node import Node
from dataspree.inference.core.producer import Producer


class Graphical(Node):

    def __init__(self, name: str, window_name: str = 'Image', classes: Optional[List] = None,
                 max_width: int = 1280, max_height: int = 720) -> None:
        super().__init__(name)

        self.consumer: Consumer = Consumer(name, maxsize=1)
        self.add_consumer(self.consumer)

        self.window_name: str = window_name
        self.classes: Optional[List] = classes
        self.max_width: int = max_width
        self.max_height: int = max_height

    def visualize(self, item: Item):
        image = item['image']
        show_image = image

        draw_img = Image.fromarray(image)

        if 'inference' in item:
            if 'detection' in item['inference'].keys():
                detections = item['inference']['detection']
                # n_detections = len(detections)

                image_pil = self.draw_boxes(draw_img, detections, color=(50, 200, 60, 255),
                                            box_labels=None,
                                            confidences=True)
            elif 'classification' in item['inference'].keys():
                image_pil = draw_img
            elif 'anomaly' in item['inference'].keys():
                image_pil = draw_img
            else:
                raise ValueError('no inference results')
        else:
            image_pil = draw_img

        show_image = np.array(image_pil)
        show_image = show_image[:, :, ::-1]
        show_image = np.array(Image.fromarray(show_image))
        aspect_ratio = image.shape[0] / image.shape[1]
        # max_side = max(image.shape[0], image.shape[1])
        if aspect_ratio >= 1:
            width = self.max_size[0]
            height = width / image.shape[0] * image.shape[1]
        else:
            height = self.max_size[1]
            width = height / image.shape[1] * image.shape[0]
        if width > self.max_size[0]:
            scaledown = self.max_size[0] / width
            height *= scaledown
            width *= scaledown
        if height > self.max_size[1]:
            scaledown = self.max_size[1] / height
            height *= scaledown
            width *= scaledown
        display_size = (int(height), int(width))
        # display_size = (image.shape[0]*2, image.shape[1]*2)
        show_image = cv2.resize(show_image, display_size)
        if 'inference' in item and 'classification' in item['inference'].keys():
            # font
            font = cv2.FONT_HERSHEY_SIMPLEX
            # org
            org = [show_image.shape[1] - 150, 1]
            # fontScale
            fontScale = 0.8
            # color in BGR
            color = (240, 231, 219)
            max_color = (100, 100, 255)
            # Line thickness of 2 px
            thickness = 2
            confidence_list = [(v) for k, v in
                               item['inference']['classification']['class_confidences'].items()]
            max_id = np.argmax(confidence_list)
            # text to display
            text = ''
            for i, key in enumerate(item['inference']['classification']['class_confidences']):
                # print(key, type(key))
                # key = str(key)
                text = str(key) + ': ' + str(
                    round(item['inference']['classification']['class_confidences'][key], 4))
                org[1] += 30
                if i == max_id:
                    color_ = max_color
                else:
                    color_ = color
                show_image = cv2.putText(show_image, text, tuple(org), font,
                                         fontScale, color_, thickness, cv2.LINE_AA)
        if 'inference' in item and 'anomaly' in item[
            'inference'].keys():  # and not 'detection' in item['inference'].keys():
            mask = item['inference']['anomaly']
            mask = cv2.resize(mask, (display_size), interpolation=cv2.INTER_NEAREST)
            ind = np.argwhere(mask == 1)
            x_dataspree_inference = ind[:, 0]
            y_dataspree_inference = ind[:, 1]

            show_image[x_dataspree_inference, y_dataspree_inference, :] = 0
            show_image[x_dataspree_inference, y_dataspree_inference, 2] = 255
            show_image[x_dataspree_inference, y_dataspree_inference, 0] = 255

        if Producer.item_route_key in item:
            item_route = item[Producer.item_route_key]
            if len(item_route) > 1:
                start = item_route[0]
                end = item_route[-1]
                dt = (end[0] - start[0]) * 1000
                text = f'Latency {dt:.0f} ms'
                font = cv2.FONT_HERSHEY_SIMPLEX
                org = (1, 15)
                font_scale = 0.5
                color = (240, 240, 240)
                thickness = 1
                show_image = cv2.putText(show_image, text, org, font,
                                         font_scale, color, thickness, cv2.LINE_AA)

        # show_image = cv2.cvtColor(show_image, cv2.COLOR_RGB2BGR)
        cv2.imshow('image', show_image.astype(np.uint8))

        cv2.waitKey(10)

    def run(self):
        while self.running:
            item = self.consumer.receive()
            if item is None:
                time.sleep(0.001)
                continue
            self.visualize(item)

        self.consumer.disconnect_all_producers()

    def draw_boxes(self, image, detections, color=(255, 255, 255, 128), box_labels=None,
                   confidences=None):
        colors = [
            (221, 221, 221, 255),
            (230, 25, 75, 255),
            (60, 180, 75, 255),
            (255, 225, 25, 255),
            (67, 99, 216, 255),
            (245, 130, 49, 255),
            (145, 30, 180, 255),
            (66, 212, 244, 255),
            (240, 50, 230, 255),
            (191, 239, 69, 255),
            (250, 190, 190, 255),
            (70, 153, 144, 255),
            (230, 190, 255, 255),
            (154, 99, 36, 255),
            (255, 250, 200, 255),
            (128, 0, 0, 255),
            (170, 255, 195, 255),
            (128, 128, 0, 255),
            (255, 216, 177, 255),
            (0, 0, 117, 255),
            (169, 169, 169, 255)
        ]

        for i in range(len(detections)):
            d = detections[i]

            if self.classes is not None:
                # if d.get('label') not in [self.classes[cl] for cl in self.classes.keys()]
                if str(d.get('label', 0)) not in self.classes:
                    continue

            draw = ImageDraw.Draw(image, 'RGBA')

            image_width = image.width
            image_height = image.height

            x = d['x'] * image_width
            y = d['y'] * image_height
            l = d['width'] * image_width
            w = d['height'] * image_height

            angle = d.get('orientation', 0.0)

            p1 = np.array([-0.5 * l, +0.5 * w])
            p2 = np.array([-0.5 * l, -0.5 * w])
            p3 = np.array([+0.5 * l, -0.5 * w])
            p4 = np.array([+0.5 * l, +0.5 * w])

            R = np.array([[np.cos(angle), -np.sin(angle)], [np.sin(angle), np.cos(angle)]])

            p1 = R.dot(p1)
            p2 = R.dot(p2)
            p3 = R.dot(p3)
            p4 = R.dot(p4)

            p1 += np.array([x, y])
            p2 += np.array([x, y])
            p3 += np.array([x, y])
            p4 += np.array([x, y])

            c_idx = int(d.get('label', 0)) % len(colors)
            color = colors[c_idx]
            # color = colors[0]

            draw.line([tuple(p1), tuple(p2), tuple(p3), tuple(p4), tuple(p1)], width=4, fill=color)

        return image

    @classmethod
    def get_config_entries(cls) -> Dict:
        return dict(display_name='Graphical',
                    updatable=True,
                    description='Opens a window on the core that renders the video for debugging purposes.',
                    fields=[
                        dict(name='window_name',
                             display_name='Window Name',
                             type=ConfigFieldTypes.STRING_INPUT,
                             default='Image'),
                        dict(name='max_width',
                             display_name='Max. width ',
                             description='Maximum width of the window.',
                             type=ConfigFieldTypes.INTEGER_INPUT,
                             default=1280),
                        dict(name='max_height',
                             display_name='Max. height ',
                             description='Maximum height of the window.',
                             type=ConfigFieldTypes.INTEGER_INPUT,
                             default=720)])

    @staticmethod
    def get_config_categories() -> List[str]:
        return ['outputs']
