import copy
import time
import cv2
import logging
import matplotlib.pyplot as plt
import numpy as np
from typing import List, Tuple, Optional, Dict, Union, Set
from abc import ABC, abstractmethod

from matplotlib import pyplot as plt
import matplotlib

from dataspree.inference.plugins.tracking.sigma_point_integration import compute_sigma_point_integral_poi, \
    compute_sigma_point_integral_map
from dataspree.inference.plugins.tracking.assignment import Murphy as Optimizer
from dataspree.inference.plugins.tracking.gaussian import draw_covariance_ellises, log_evaluate_gaussians, log_evaluate_gaussians_2, \
    evaluate_gaussians

logger = logging.getLogger(__name__)

min_accepted = 4e-2  # 2e-60  2e-80
# matplotlib.use('agg')

"""
Parameters for the configuration

- p_c small p_c  -> probability of assigning a new detection to an existing obejct grows
- p_d: small p_d -> the likelihood of not assigning a detection grows (can be counteracted by lowering p_c)
                 -> objects die later


"""

# List of distinct colors that are suitable for visualizing pig trajectories
DISTINCT_TRACKING_COLORS: List[Tuple[int, int, int]] = [
    (47, 79, 79), (85, 107, 47), (139, 69, 19),
    (107, 142, 35), (160, 82, 45), (165, 42, 42), (46, 139, 87), (34, 139, 34), (127, 0, 0), (25, 25, 112), (0, 100, 0),
    (112, 128, 144), (139, 0, 0), (128, 128, 0), (72, 61, 139), (178, 34, 34), (95, 158, 160), (0, 128, 0),
    (60, 179, 113), (102, 51, 153), (0, 139, 139), (70, 130, 180), (0, 0, 128), (210, 105, 30), (154, 205, 50),
    (32, 178, 170), (75, 0, 130), (50, 205, 50), (218, 165, 32), (127, 0, 127), (143, 188, 143), (139, 0, 139),
    (176, 48, 96), (72, 209, 204), (102, 205, 170), (153, 50, 204), (255, 0, 0), (255, 69, 0), (0, 206, 209),
    (255, 140, 0), (255, 165, 0), (255, 215, 0), (106, 90, 205), (255, 255, 0), (199, 21, 133), (0, 0, 205),
    (64, 224, 208), (127, 255, 0), (0, 255, 0), (148, 0, 211), (186, 85, 211), (0, 250, 154), (138, 43, 226),
    (0, 255, 127), (65, 105, 225), (220, 20, 60), (0, 255, 255), (0, 191, 255), (244, 164, 96), (147, 112, 219),
    (0, 0, 255), (160, 32, 240), (173, 255, 47), (255, 99, 71), (218, 112, 214), (255, 0, 255), (30, 144, 255),
    (240, 230, 140), (238, 232, 170), (255, 255, 84), (100, 149, 237), (221, 160, 221), (176, 224, 230),
    (144, 238, 144), (135, 206, 235), (123, 104, 238), (238, 130, 238), (135, 206, 250), (127, 255, 212)]

logger = logging.getLogger(__name__)


def get_track_color(tid: np.ndarray, as_float: bool = False) -> Union[Tuple[float, float, float],
                                                                      Tuple[int, int, int, int]]:
    color = DISTINCT_TRACKING_COLORS[hash(tuple(tid)) % len(DISTINCT_TRACKING_COLORS)]
    if as_float:
        color = color[0] / 255.0, color[1] / 255.0, color[2] / 255.0

    return color


def rotate_mat(points: np.ndarray, rads: np.ndarray) -> None:
    number_of_rotations: int = points.shape[0]
    number_of_points_per_rotation: int = points.shape[1]
    point_dimensionality: int = 2
    assert points.shape == (number_of_rotations, number_of_points_per_rotation, point_dimensionality)
    assert rads.shape == (number_of_rotations,)

    rotation_matrix = np.array([[np.cos(rads), -np.sin(rads)], [np.sin(rads), np.cos(rads)]])
    assert rotation_matrix.shape == (point_dimensionality, point_dimensionality, number_of_rotations)

    for i, ps in enumerate(points):
        points[i] = rotation_matrix[:, :, i] @ ps @ rotation_matrix[:, :, i].T

    # np.einsum('dDn,npd,dDn->npd', rotation_matrix, points, rotation_matrix, out=points)
    # assert points.shape == (number_of_rotations, number_of_points_per_rotation, point_dimensionality)


def rotate(points: np.ndarray, rads: np.ndarray) -> None:
    if __debug__:
        __points = points.copy()

    number_of_rotations: int = points.shape[0]
    number_of_points_per_rotation: int = points.shape[1]
    point_dimensionality: int = 2
    assert points.shape == (number_of_rotations, number_of_points_per_rotation, point_dimensionality)
    assert rads.shape == (number_of_rotations,)

    rotation_matrix = np.array([[np.cos(rads), -np.sin(rads)], [np.sin(rads), np.cos(rads)]])
    assert rotation_matrix.shape == (point_dimensionality, point_dimensionality, number_of_rotations)

    np.einsum('Ddn,npd->npD', rotation_matrix, points, out=points)
    assert points.shape == (number_of_rotations, number_of_points_per_rotation, point_dimensionality)

    if __debug__:
        for i_n_r in range(number_of_rotations):
            rad = rads[i_n_r]
            for i_n_p in range(number_of_points_per_rotation):
                point = __points[i_n_r][i_n_p]
                __rotated_point = np.array([[np.cos(rad), -np.sin(rad)], [np.sin(rad), np.cos(rad)]]).dot(point)
                assert (np.abs(__rotated_point - points[i_n_r, i_n_p]) < 1e-9).all()


class SurvivalProbability(ABC):

    def __init__(self, fov: Tuple[int, int, int, int]):
        self.fov: Tuple[int, int, int, int] = fov
        # Survival Probability evaluated at dummy Normal distributions with very narrow variances
        self.__p_s: Optional[np.ndarray] = None

    @abstractmethod
    def predict(self, means: np.ndarray, covariances: np.ndarray) -> np.ndarray:
        """
        Predict the probability of survival; i.e., the factor with which the object(modelled via a probability
        distribution)'s probability of existence is multiplied.

        :param means:          The means of the N probability distributions (shape N x d)
        :param covariances:    The covariance matrices of the N probability distributions (shape N x d x d)
        :return:               The probability of existence after modelling survival of objects (shape N).
        """
        ...

    def evaluate(self, image_size: Tuple[int, int], original_image_size: Optional[Tuple[int, int]] = None):
        if self.__p_s is None:
            means = np.moveaxis(np.array(np.meshgrid(np.linspace(1, original_image_size[1], image_size[1]),
                                                     np.linspace(0, original_image_size[0], image_size[0]))), 0, 2)
            means_shape = means.shape
            means = means.reshape(-1, 2)

            covariances = np.repeat(np.eye(2)[np.newaxis], means.shape[0], axis=0)

            color_bar_width = 10
            p_s = self.predict(means=means, covariances=covariances).reshape(means_shape[:2])
            p_s = np.hstack((p_s, np.empty((p_s.shape[0], color_bar_width))))
            p_s[:, -color_bar_width:] = np.linspace(0, 1, p_s.shape[0])[:, np.newaxis]

            p_s = cv2.applyColorMap((p_s * 255).astype(np.uint8), cv2.COLORMAP_HOT)
            self.__p_s = p_s

        return self.__p_s.copy()


class RoiSurvival(SurvivalProbability):
    """
    Compute the probability of survival assuming a (configurable) constant density inside a field of view and 0
    outside.
    """

    def __init__(self, p_s: float, fov: Tuple[int, int, int, int]):
        """
        :param p_s:               Probability density assumed inside the #fov
        :param fov:               Field of view inside which the density is set to #p_s
        """
        super().__init__(fov=fov)
        self.p_s: float = p_s
        self.fov_integrate: Tuple[float, float, float, float] = (float(fov[0]), float(fov[1]),
                                                                 float(fov[2]), float(fov[3]))

    def predict(self, means: np.ndarray, covariances: np.ndarray) -> np.ndarray:
        """
        :see SurvivalProbability.predict
        """
        number_of_elements: int = means.shape[0]
        state_dimensionality: int = means.shape[1]
        assert means.shape == (number_of_elements, state_dimensionality)
        assert covariances.shape == (number_of_elements, state_dimensionality, state_dimensionality)

        return compute_sigma_point_integral_poi(means=means, covariances=covariances, roi=self.fov_integrate,
                                                p_d=self.p_s)


class ReducedRoiSurvival(RoiSurvival):
    """
    Reduces the field of view in x and y dimensions (symmetrically at start and end of an axis).
    """

    def __init__(self, p_s: float, fov: Tuple[int, int, int, int], horizontal_area_retained: float = 1.0,
                 vertical_area_retained: float = 1.0) -> None:
        """

        :param p_s:  @see RoiSurvival.__init__
        :param fov:  @see RoiSurvival.__init__
        :param horizontal_area_retained:  The percentage of the area in the x dimension that is retained
        :param vertical_area_retained:    The percentage of the area in the y dimension that is retained
        """
        super().__init__(p_s=p_s, fov=fov)

        diam_reduction_horizontal: float = (fov[3] - fov[1]) * (1 - horizontal_area_retained) / 2
        diam_reduction_vertical: float = (fov[2] - fov[0]) * (1 - vertical_area_retained) / 2
        self.fov_integrate: Tuple[float, float, float, float] = (fov[0] + diam_reduction_vertical,
                                                                 fov[1] + diam_reduction_horizontal,
                                                                 fov[2] - diam_reduction_vertical,
                                                                 fov[3] - diam_reduction_horizontal)


class SmoothTransitionSurvival(SurvivalProbability):
    """
    Compute the probability of survival as a smooth (and linear) transition between the borders and the field of view's
    center.
    """

    def __init__(self, p_s: float, fov: Tuple[int, int, int, int], horizontal_area_retained: float = 1.0,
                 vertical_area_retained: float = 1.0, horizontal_percentage_zero: float = 0.25,
                 vertical_percentage_zero: float = 0.25, min_value: float = 0.0) -> None:
        """
        :param p_s:                       Probability density assumed inside the #fov (and with which every value in
                                          the pmf is multiplied).
        :param fov:                       Field of view outside which the density is set to 0.
        :param horizontal_area_retained:  The percentage of the area in the x dimension that is not part of the
                                          transition, i.e. inside which the probability of survival equals #p_s.
        :param vertical_area_retained:    The percentage of the area in the x dimension that is not part of the
                                          transition, i.e. inside which the probability of survival equals #p_s.
        """
        super().__init__(fov=fov)
        self.p_s: float = p_s
        assert fov[0] == 0 and fov[1] == 0

        row_prc_start: float = (1 - vertical_area_retained) / 2
        col_prc_start: float = (1 - horizontal_area_retained) / 2

        # Set Linear falloff at borders
        self.g_map = np.empty(fov[2:])
        rows, cols = self.g_map.shape
        for row in range(rows):
            row_prc = min(row / rows, (rows - row) / rows)
            vertical_factor = 1 if row_prc_start == 0 \
                else min(1.0, max(0.0,
                                  (row_prc - row_prc_start * vertical_percentage_zero)
                                  / (row_prc_start * (1 - vertical_percentage_zero))))

            for col in range(cols):
                col_prc = min(col / cols, (cols - col) / cols)
                horizontal_factor = 1 if col_prc_start == 0 \
                    else min(1.0, max(0.0, (col_prc - col_prc_start * horizontal_percentage_zero)
                                      / (col_prc_start * (1 - horizontal_percentage_zero))))

                self.g_map[row, col] = (horizontal_factor * vertical_factor * p_s) * (1 - min_value) + p_s * min_value

        # Set borders to zero because the sigma point integral clips lookup indices at the borders
        self.g_map[[0, -1]] = 0
        self.g_map[:, [0, -1]] = 0

    def predict(self, means: np.ndarray, covariances: np.ndarray) -> np.ndarray:
        """
        :see SurvivalProbability.predict
        """
        number_of_elements: int = means.shape[0]
        state_dimensionality: int = means.shape[1]
        assert means.shape == (number_of_elements, state_dimensionality)
        assert covariances.shape == (number_of_elements, state_dimensionality, state_dimensionality)

        spi, poi = compute_sigma_point_integral_map(means=means, covariances=covariances, g_map=self.g_map,
                                                    return_poi=True)

        def show():
            print('...')
            for i in range(spi.shape[0]):
                points = poi[:, i, :2]
                plt.scatter(*points.T, label=str(i))

            for i, f in enumerate(spi):
                print(i, f, means)
            plt.scatter([self.fov[0], self.fov[2]], [self.fov[1], self.fov[3]])
            plt.legend()
            plt.show()

        return spi


class LinearMotionModel:

    @staticmethod
    def create_random_walk(dimensionality: int) -> 'LinearMotionModel':
        return LinearMotionModel(motion_model=np.eye(dimensionality))

    @staticmethod
    def create_walk(dimensionality: int, speed_time_factor: float = 0.1, no_speed_factor=1) -> 'LinearMotionModel':
        motion_model = np.eye(dimensionality)
        v = no_speed_factor
        # motion_model[-2:, :2] = np.array([[v, -v], [-v, v]])
        motion_model[-2:, :2] = np.array([[v, 0], [0, v]])

        motion_model_time = np.eye(dimensionality, k=-2) * speed_time_factor
        # motion_model_time[-1, 0] = -speed_time_factor
        # motion_model_time[-2, 1] = -speed_time_factor

        return LinearMotionModel(motion_model=motion_model, motion_model_time=motion_model_time)

    def __init__(self, motion_model: np.ndarray, motion_model_time: Optional[np.ndarray] = None):
        self.motion_model = motion_model
        self.motion_model_time = motion_model_time

    def predict(self, dt: float) -> np.ndarray:
        if self.motion_model_time is None:
            return self.motion_model.copy()

        return self.motion_model + dt * self.motion_model_time


class GaussianTransitionDensity:

    def __init__(self, linear_motion_model: LinearMotionModel,
                 noise_covariance: LinearMotionModel,
                 survival_probability: SurvivalProbability,
                 state_orientation_column: Optional[int], rotate_noise_covariance: bool) -> None:
        self.linear_motion_model: LinearMotionModel = linear_motion_model
        self.noise_covariance: LinearMotionModel = noise_covariance

        self.survival_probability: SurvivalProbability = survival_probability
        self.rotate_noise_covariance: bool = rotate_noise_covariance
        self.state_orientation_column: Optional[int] = state_orientation_column

    def predict(self, probabilities_of_existence: np.ndarray, means: np.ndarray, covariances: np.ndarray,
                dt: float) -> None:
        """
        Perform a prediction with F := linear_motion_model
                                  Q := noise_covariance

        means                       :=  F means
        covariances                 :=  F covariances F^T  + Q

        probabilities_of_existence  :*= \\E_{N( ; means, covariances)} [ \\1_{\\in roi} * self.survival_probability ]


        :param probabilities_of_existence:      Current probabilities of existence (shape N)
        :param means:                           Current means (shape N x d)
        :param covariances:                     Current covariance matrices (shape N x d x d)
        :param dt:                              Time difference between current time and last time

        :return: (probabilities_of_existence, means, covariances)
        """
        if __debug__:
            __r = probabilities_of_existence.copy()
            __means = means.copy()
            __covariance_matrices = covariances.copy()

        number_of_objects: int = probabilities_of_existence.shape[0]
        assert probabilities_of_existence.shape == (number_of_objects,)

        state_dimensionality: int = means.shape[1]
        assert means.shape == (number_of_objects, state_dimensionality)
        assert covariances.shape == (number_of_objects, state_dimensionality, state_dimensionality)

        # Compute each object's probability of survival
        p_s_bar = self.survival_probability.predict(means=means, covariances=covariances)
        assert p_s_bar.shape == (number_of_objects,)
        np.maximum(p_s_bar, min_accepted, out=p_s_bar)

        # Update the object's probability of existence by multiplying it with the probability of survival evaluated
        # at the object's position.
        np.multiply(probabilities_of_existence, p_s_bar, out=probabilities_of_existence)
        assert probabilities_of_existence.shape == (number_of_objects,)

        # compute the motion model matrix F that is applied to the mean and covariance
        motion_model_matrix = self.linear_motion_model.predict(dt)
        assert motion_model_matrix.shape == (state_dimensionality, state_dimensionality)

        # F \\cdot means[i]
        np.einsum('Dd,nd->nD', motion_model_matrix, means, out=means)
        assert means.shape == (number_of_objects, state_dimensionality)

        noise_covariance = self.noise_covariance.predict(dt)
        if self.state_orientation_column is not None and self.rotate_noise_covariance and self.state_orientation_column:
            noise_covariance = np.repeat(noise_covariance[None], number_of_objects, axis=0)
            rotate_mat(noise_covariance[:, :2, :2], -means[:, 4])

        # F \\cdot covariances \\cdot F.T + Q
        np.einsum('ab,nbc,ec->nae', motion_model_matrix, covariances, motion_model_matrix, out=covariances)
        np.add(covariances, noise_covariance, out=covariances)
        assert covariances.shape == (number_of_objects, state_dimensionality, state_dimensionality)

        if __debug__:
            for i in range(number_of_objects):
                __alternative_mean = motion_model_matrix.dot(__means[i])
                __diff = np.abs(__alternative_mean - means[i])
                assert (1e-9 > __diff).all()

                __alternative_r = p_s_bar[i] * __r[i]
                __diff = np.abs(probabilities_of_existence[i] - __alternative_r)
                assert 1e-9 > __diff

            for i, cov in enumerate(__covariance_matrices):
                __alternative_cov = motion_model_matrix.dot(cov).dot(motion_model_matrix.T) + noise_covariance[i]
                __diff = np.abs(__alternative_cov - covariances[i])
                assert (1e-9 > __diff).all()


class MultibernoulliBirthModel:
    """



    Part of the predicted, undetected parametrization

    """

    @staticmethod
    def create(fov: Tuple[int, int, int, int],
               position_uncertainty: float,
               gaussians_per_side: int = 1,
               # velocity_mean: Optional[float] = None, velocity_variance: Optional[float] = None,
               # velocity_weigth=10,
               # angle_variance: Optional[float] = None,
               # confidence_mean: Optional[float] = None, confidence_variance: Optional[float] = None,
               birth_weight_scale: float = min_accepted,  # 0.000002,
               horizontal: bool = True,
               additional_mean: Optional[np.ndarray] = None,
               additional_cov: Optional[np.ndarray] = None) -> 'MultibernoulliBirthModel':
        # the birth weights for the PPP don't have to sum to one, they indicate the expected number of born
        # objects per frame
        birth_weights = np.ones(gaussians_per_side * 2) * birth_weight_scale

        offset_width, offset_height = fov[0], fov[1]
        image_width, image_height = fov[2] - fov[0], fov[3] - fov[1]

        position_variance = (image_height * position_uncertainty) ** 2 / gaussians_per_side
        position_std_dev = np.sqrt(position_variance)

        std_dev_factor = 1.0

        prc_total = 0.6  # percentage covered with gaussians
        if horizontal:

            birth_means = np.array(
                [[offset_width + ((1 - prc_total) / 2) * image_width + image_width * prc_total / (
                        gaussians_per_side + 1) * (i + 1), position_std_dev * std_dev_factor]
                 for i in
                 range(gaussians_per_side)])

            swap_index = len(birth_means)

            birth_means = np.vstack((birth_means,
                                     [[offset_width + (1 - prc_total) / 2 * image_width + image_width * prc_total / (
                                             gaussians_per_side + 1) * (i + 1),
                                       offset_height + image_height - position_std_dev * std_dev_factor] for i in
                                      range(gaussians_per_side)]
                                     ))

            birth_cov = np.array(
                [
                    # [(image_width / 2) ** 2, 0],
                    [position_variance, 0],
                    [0, position_variance]
                ])
        else:
            birth_means = np.array(
                [[position_std_dev * std_dev_factor,
                  offset_height + image_height * (1 - prc_total) / 2 + image_height * prc_total / (
                          gaussians_per_side + 1) * (i + 1)]
                 for i in
                 range(gaussians_per_side)])

            swap_index = len(birth_means)

            birth_means = np.vstack((birth_means,
                                     [[offset_width + image_width - position_std_dev * std_dev_factor,
                                       offset_height + image_height * (1 - prc_total) / 2 + image_height * prc_total / (
                                               gaussians_per_side + 1) * (i + 1)] for i in
                                      range(gaussians_per_side)]
                                     ))

            birth_cov = np.array(
                [
                    [position_variance, 0],
                    [0, position_variance]
                ])

        if additional_mean is not None or additional_cov is not None:
            assert (additional_mean is None) == (additional_cov is None)
            number_of_new_dimensions = additional_mean.shape[0]

            assert additional_mean.shape == (number_of_new_dimensions,)
            assert additional_cov.shape == (number_of_new_dimensions, number_of_new_dimensions)  # no cov with xy

            birth_means = np.pad(birth_means, ((0, 0), (0, number_of_new_dimensions)))
            birth_means[:, -number_of_new_dimensions:] = additional_mean

            birth_cov = np.pad(birth_cov, (0, number_of_new_dimensions))
            birth_cov[-number_of_new_dimensions:, -number_of_new_dimensions:] = additional_cov

        """
        if confidence_mean is not None:
            assert confidence_variance is not None

            birth_means = np.pad(birth_means, ((0, 0), (0, 1)))
            birth_means[-1] = confidence_mean

            birth_cov = np.pad(birth_cov, (0, 1))
            birth_cov[-1, -1] = confidence_variance  # make this influence the other values

        if velocity_mean is not None:
            velocity_variance = velocity_variance if velocity_variance is not None else velocity_mean

            birth_means = np.pad(birth_means, ((0, 0), (0, 2)))
            birth_means[0: swap_index, -1] = velocity_mean
            birth_means[swap_index:, -1] = -velocity_mean

            birth_cov = np.pad(birth_cov, (0, 2))
            birth_cov[-2, -2] = velocity_variance
            birth_cov[-1, -1] = velocity_variance

            birth_cov[0, -2] = velocity_weigth
            birth_cov[1, -1] = velocity_weigth

        if angle_variance is not None:
            angle_mean = np.pi / 2  # erro
            birth_means = np.pad(birth_means, ((0, 0), (0, 2)))
            birth_means[0: swap_index, -1] = angle_mean
            birth_means[swap_index:, -1] = -angle_mean  # err not negative

            birth_cov = np.pad(birth_cov, (0, 2))
            birth_cov[4, 4] = angle_variance
            birth_cov[5, 5] = angle_variance
        """

        return MultibernoulliBirthModel(w=birth_weights, mu=birth_means, cov=birth_cov)

    def __init__(self, w: np.ndarray, mu, cov):
        """
        w: weights
        mu: means of the gaussians
        cov: covariance matrix (P) of the birth model.
        """

        if len(cov.shape) == 2:
            cov = cov[np.newaxis]
            co = cov.copy()
            for _ in range(w.shape[0] - 1):
                cov = np.vstack((cov, co))

            assert len(cov.shape) == 3, f'{len(cov.shape)} {cov.shape}'

        self.inv = None
        self.w = w
        self.mu = mu
        self.cov = cov

        self.base_labels = np.array(list(zip((0,) * self.get_number_of_birth_components(),
                                             range(self.get_number_of_birth_components()))))

        assert self.base_labels.shape[0] == self.get_number_of_birth_components()
        assert self.base_labels.shape[1] == 2

        assert len(w.shape) == 1
        assert len(mu.shape) == 2
        assert len(self.cov.shape) == 3, f'{self.cov.shape}'

        self.inv = None
        number_of_birth_components: int = self.get_number_of_birth_components()
        assert number_of_birth_components == mu.shape[0]
        assert number_of_birth_components == cov.shape[0], f'{number_of_birth_components} {cov.shape}'
        assert number_of_birth_components == w.shape[0]

        state_dimensionality: int = self.get_state_dimensionality()
        assert state_dimensionality == mu.shape[1]
        assert state_dimensionality == cov.shape[1]
        assert state_dimensionality == cov.shape[2]

    @staticmethod
    def evaluate(x, ws, mus, invs, divisors):
        su = None

        for i, m in enumerate(mus):
            diff = x - m
            eid = np.sum(np.dot(invs[i], diff.T).T * diff, axis=1)
            val = ws[i] * np.exp(-1 / 2. * eid) / divisors[i]
            if su is None:
                su = val
            else:
                su += val

        return su

    def __call__(self, x: np.ndarray):
        if self.inv is None:
            self.inv = np.linalg.inv(self.cov)
            self.divisor = np.sqrt(2 * np.pi ** self.get_state_dimensionality() * np.linalg.det(self.cov))

        return self.evaluate(x, self.w, self.mu, self.inv, self.divisor)

    def __setattr__(self, name, y):
        if name == 'cov':
            self.inv = None

        return super().__setattr__(name, y)

    def get_number_of_birth_components(self) -> int:
        return self.w.shape[0]

    def get_state_dimensionality(self) -> int:
        return self.mu.shape[1]

    def predict(self, time_count: int):
        return self.base_labels + (time_count, 0), self.w.copy(), self.mu.copy(), self.cov.copy()

    def predict_put(self, time_count: int, *, labels: np.ndarray, weights: np.ndarray, means: np.ndarray,
                    covariance_matrices: np.ndarray):
        number_of_birth_components: int = self.get_number_of_birth_components()
        labels[:number_of_birth_components] = self.base_labels + (time_count, 0)
        weights[:number_of_birth_components] = self.w
        means[:number_of_birth_components] = self.mu
        covariance_matrices[:number_of_birth_components] = self.cov


class GaussianPredictionParametrization:
    def __init__(self, birth: MultibernoulliBirthModel, transition: GaussianTransitionDensity):
        self.birth: MultibernoulliBirthModel = birth
        self.transition: GaussianTransitionDensity = transition

    def get_information(self):
        return dict(birth_w=self.birth.w, birth_mu=self.birth.mu, birth_cov=self.birth.cov,
                    transition=self.transition.linear_motion_model.motion_model,
                    transition_t=self.transition.linear_motion_model.motion_model_time,
                    survival_probability=self.transition.survival_probability.p_s,
                    noise_covariance=self.transition.noise_covariance)


class UpdateModel:

    @staticmethod
    def create(dimensionality: int, var: Union[float, List[float]] = 30.0, p_d: float = 0.8,
               p_c: float = 0.05 / 640 / 420) -> 'UpdateModel':
        return UpdateModel(transform=np.eye(dimensionality), covariance=np.eye(dimensionality) * var, p_d=p_d, p_c=p_c)

    def __init__(self, transform: np.ndarray, covariance: np.ndarray, p_d: float = 0.8, p_c: float = 0.05 / 640 / 420):
        self.transform: np.ndarray = transform
        self.covariance: np.ndarray = covariance

        self.p_d: float = p_d
        self.p_c: float = p_c

    def get_information(self):
        return dict(p_d=self.p_d, p_c=self.p_c,
                    transformation_matrix=self.transform,
                    transformation_covariance=self.covariance)


class MultibernoulliMixture:
    """
    Track an unknown and time-varying number of objects.
    This implementation is Hypothesis oriented.

    MBM( x ) = \\sum_h w^h MB^h(x)
             = \\sum_h w^h \\sum_{\\dis_union x^j = x} \\prod_i B^{i;h}(x)

    with a bernoulli distribution B = B(x;r, P) = 1 - r  if len(x) == 0
                                                = r P(x[0])    if len(x) == 1
                                                = 0      else
    in this case implemented with a Gaussian conjugate prior P = P(x; \\mu, \\Sigma) = N(x ; \\mu, \\sigma)

    Semantic of variables / parameters
        - \\mu mean of the single-object distribution
        - \\Sigma covariance matrix of the single-object distribution
        - r probability of existence


    Open small tasks:
    - Give the possibility to increase the clutter probability if a sample's confidence is very low ()


    Open tasks that can be performed if necessary:
    - Implement PMBM ( would possibly be an improvement for venneker, as this remvoes the possiblitliy of a pig being
                       misdetected along the edge )
    - Implement Track Oriented MBM and use it in case the number of hypothesis is larger than one.

    """

    def __init__(self, predict: GaussianPredictionParametrization,
                 update: UpdateModel, max_number_of_hypothesis: int,
                 prune_weight: float = 0.001, object_prune_probability: float = min_accepted,
                 object_exists_probability: float = min_accepted,  # * 1e3 # 1e-15
                 horizontal: bool = True,
                 detection_size_columns: Optional[Tuple[int, int]] = (2, 3),
                 detection_orientation_column: Optional[int] = 4,
                 detection_confidence_column: Optional[int] = 5,
                 detection_size_divided_by: float = 1,
                 state_size_columns: Optional[Tuple[int, int]] = (2, 3),
                 state_orientation_column: Optional[int] = 4) -> None:
        assert horizontal
        self.horizontal: bool = horizontal
        self.predict: GaussianPredictionParametrization = predict
        self.update: UpdateModel = update

        self.prune_log_weight: float = np.log(prune_weight)
        self.object_prune_probability: float = object_prune_probability
        self.object_exists_probability: float = object_exists_probability
        self.max_number_of_hypothesis: int = max_number_of_hypothesis

        # L
        self.log_weights = np.empty(max_number_of_hypothesis)

        self.lbl: List[np.ndarray] = []
        self.r: List[np.ndarray] = []

        # p^h, i
        self.mean: List[np.ndarray] = []
        self.cov: List[np.ndarray] = []
        self.time_count = 0

        # (object_identifier => [means, current probability of existence, color, number_of_detections ])
        self.tracks: Dict[Tuple[int, int], Tuple[np.ndarray, float, Tuple[int, int, int]], int] = dict()
        self.count_out: int = 0
        self.count_in: int = 0

        self.print_information()
        self.plot: bool = False

        self.__used_colors: Dict[Tuple[int, int], Tuple[int, int, int]] = dict()
        self.__colors: Set[Tuple[int, int, int]] = set(DISTINCT_TRACKING_COLORS)

        # Semantic of the detection array
        self.detection_size_columns: Optional[Tuple[int, int]] = detection_size_columns
        self.detection_orientation_column: Optional[int] = detection_orientation_column
        self.detection_confidence_column: Optional[int] = detection_confidence_column

        self.detection_size_divided_by: float = detection_size_divided_by

        self.state_size_columns: Optional[Tuple[int, int]] = state_size_columns
        self.state_orientation_column: Optional[int] = state_orientation_column

        self.__last_visualization = None

        self.__to = np.empty((100, 4), dtype=np.float)
        self.__i = 0

        self.__times = np.empty((30000, 8), dtype=np.float)
        self.__times_i = 0
        self.__visualize_fully: bool = False

    def reset(self) -> Tuple[int, int, int, int]:

        new_in, new_out, counted_objects_lifetime = self.__count(dict())
        total_in, total_out = self.count_in, self.count_out

        self.count_out = 0
        self.count_in = 0
        self.tracks = dict()
        self.lbl = []
        self.r = []
        self.mean = []
        self.cov = []
        self.time_count = 0

        return new_in, new_out, total_in, total_out, counted_objects_lifetime

    def print_information(self):

        prediction_configuration = self.predict.get_information()
        update_configuration = self.update.get_information()
        pruning_configuration = dict(prune_weights=np.exp(self.prune_log_weight),
                                     object_prune_probability=self.object_prune_probability,
                                     object_exists_probability=self.object_exists_probability,
                                     max_number_of_hypothesis=self.max_number_of_hypothesis)

        print('\nprediction')
        for k, v in prediction_configuration.items():
            print(f'->{k}\n:{v}')

        print('\nupdate configuration')
        for k, v in update_configuration.items():
            print(f'->{k}\n:{v}')

        print('\npruning configuration')
        for k, v in pruning_configuration.items():
            print(f'->{k}\n:{v}')

    def tic(self, dt: float, z: np.ndarray, confidences: Optional[np.ndarray] = None, *,
            visualization_image: Optional[np.ndarray] = None, all_detections: Optional[np.ndarray] = None):

        switch_visu: bool = True
        _times_i: int = self.__times_i % self.__times.shape[0]
        self.__times[_times_i][0] = time.time()

        # Prior state
        __states = []
        if visualization_image is not None and switch_visu:
            __states.append([copy.deepcopy(self.log_weights), copy.deepcopy(self.lbl), copy.deepcopy(self.r),
                             copy.deepcopy(self.mean), copy.deepcopy(self.cov)])

        # Prediction
        self.__predict(dt)
        self.__times[_times_i][1] = time.time()
        if visualization_image is not None and switch_visu:
            __states.append([copy.deepcopy(self.log_weights), copy.deepcopy(self.lbl), copy.deepcopy(self.r),
                             copy.deepcopy(self.mean), copy.deepcopy(self.cov)])

        # Update
        self.__update(z, confidences=confidences)
        self.__times[_times_i][2] = time.time()
        if visualization_image is not None and switch_visu:
            __states.append([copy.deepcopy(self.log_weights), copy.deepcopy(self.lbl), copy.deepcopy(self.r),
                             copy.deepcopy(self.mean), copy.deepcopy(self.cov)])

        # Reduction (pruning, (merging), capping)
        dead_labels: List[np.ndarray] = self.__reduce()
        self.__times[_times_i][3] = time.time()
        if visualization_image is not None and switch_visu:
            __states.append([copy.deepcopy(self.log_weights), copy.deepcopy(self.lbl), copy.deepcopy(self.r),
                             copy.deepcopy(self.mean), copy.deepcopy(self.cov)])

        new_estimates = self.__estimate()
        self.__times[_times_i][4] = time.time()

        new_in, new_out, counted_object_lifetimes = self.__count(new_estimates)
        self.__times[_times_i][5] = time.time()

        self.time_count += 1
        adapted_image = self.__visualize_tic(visualization_image, __states, all_detections=all_detections)
        # adapted_image = visualization_image
        self.__times[_times_i][6] = time.time()

        for label in dead_labels:
            tid = tuple(label.astype(np.int))
            if tid in self.__used_colors:
                self.__colors.add(self.__used_colors[tid])
                del self.__used_colors[tid]

        self.__times[_times_i][7] = time.time()

        total_duration = self.__times[_times_i][-1] - self.__times[_times_i][0]

        self.__times[_times_i][1:] -= self.__times[_times_i][:-1]
        self.__times_i += 1

        if self.__times_i % 100 == 0:
            arr = self.__times[:self.__times_i + 1]
            total = np.average(np.sum(arr[:, 1:], axis=1), axis=0)
            combined = list(zip(('predict', 'update', 'reduce', 'estimate', 'count', 'visualize', 'color'),
                                np.average(arr[:, 1:], axis=0)))
            logger.info(', '.join([f'{t}:{f:0.3f}' for t, f in combined])
                        + f', {total:0.3f}, {1 / total:0.1f} {arr.shape[0]}')

        return new_estimates, self.tracks.copy(), new_in, new_out, self.count_in, self.count_out, adapted_image, \
               counted_object_lifetimes

    def __predict(self, dt: float) -> None:
        """
        Variables:
        P^s  survival probability     self.predict.transition.survival_probability
        Transition density            self.predict.transition
        Multibernoulli Birth Model:   self.predict.birth


        Each Multibernoulli is predicted independently from the other Multibernoullis.
        :see GaussianTransitionDensity.predict

        Birth Multibernoullis are applied.
        :see MultibernoulliBirthModel.predict

        :param dt:  The time that has passed since the last update:

        """
        number_of_hypothesis_last_iteration: int = len(self.lbl)

        # Predict all existing hypothesis
        for h in self.active_hypothesis_ids():
            # birth of new objects
            lbl_b, r_b, mean_b, cov_b = self.predict.birth.predict(self.time_count)

            # predict surviving objects (inplace)
            self.predict.transition.predict(self.r[h], self.mean[h], self.cov[h], dt=dt)

            # Combine new object birth and surviving objects to form the new state of hypothesis h
            self.r[h] = np.hstack((r_b, self.r[h]))
            self.lbl[h] = np.vstack((lbl_b, self.lbl[h]))
            self.mean[h] = np.vstack((mean_b, self.mean[h]))
            self.cov[h] = np.vstack((cov_b, self.cov[h]))

            number_of_objects_in_hypothesis: int = self.r[h].shape[0]
            assert self.r[h].shape == (number_of_objects_in_hypothesis,)

            label_dimensionality: int = self.lbl[h].shape[1]
            assert self.lbl[h].shape == (number_of_objects_in_hypothesis, label_dimensionality)

            state_dimensionality: int = self.mean[h].shape[1]
            assert self.mean[h].shape == (number_of_objects_in_hypothesis, state_dimensionality)
            assert self.cov[h].shape == (number_of_objects_in_hypothesis, state_dimensionality, state_dimensionality)

        # If no hypothesis exists, model birth of new objects
        if number_of_hypothesis_last_iteration == 0:
            lbl_b, r_b, mean_b, cov_b = self.predict.birth.predict(self.time_count)

            self.lbl.append(lbl_b)
            self.r.append(r_b)
            self.mean.append(mean_b)
            self.cov.append(cov_b)
            self.log_weights[0] = 0  # log(1)

        # Ensure correct state spaces
        if __debug__:
            number_of_hypothesis_current_iteration: int = len(self.lbl)
            assert number_of_hypothesis_current_iteration > 0

            number_of_objects_in_hypothesis: int = self.r[0].shape[0]
            label_dimensionality: int = self.lbl[0].shape[1]
            state_dimensionality: int = self.mean[0].shape[1]
            for h in range(number_of_hypothesis_current_iteration):
                assert self.r[h].shape == (number_of_objects_in_hypothesis,)
                assert self.lbl[h].shape == (number_of_objects_in_hypothesis, label_dimensionality)
                assert self.mean[h].shape == (number_of_objects_in_hypothesis, state_dimensionality)
                assert self.cov[h].shape == (
                    number_of_objects_in_hypothesis, state_dimensionality, state_dimensionality)

    def __update(self, z, confidences: Optional[np.ndarray]):
        state_dimensionality: int = self.predict.birth.mu.shape[1]

        number_of_detections: int = z.shape[0]
        detection_dimensionality: int = z.shape[1]
        assert z.shape == (number_of_detections, detection_dimensionality)

        new_lbl: List = []
        new_r: List[np.ndarray] = []
        new_means: List[np.ndarray] = []
        new_covs: List[np.ndarray] = []
        new_log_weights: np.ndarray = np.empty(self.max_number_of_hypothesis * 100)

        ts = time.time()

        for h in self.active_hypothesis_ids():
            number_of_objects_in_hypothesis: int = len(self.lbl[h])

            # Compute the cost of assignment, and also return some useful values for updating the state.
            tc = time.time()
            cost_matrix, zh, mat_k, mat_p, diff = self.__compute_cost_matrix_for_hypothesis(
                z, hypothesis_identifier=h, confidences=confidences)
            tc2 = time.time()
            # print(f'{tc2 - tc:.3f}', 'tc')

            # Cost matrix of assignment and misdetection
            assert cost_matrix.shape == (number_of_objects_in_hypothesis,
                                         number_of_detections + number_of_objects_in_hypothesis)

            # Objects transformed into measurement space
            assert zh.shape == (number_of_objects_in_hypothesis, detection_dimensionality)

            # Kalman Gain
            assert mat_k.shape == (number_of_objects_in_hypothesis, state_dimensionality, detection_dimensionality)

            # New covariance matrices
            assert mat_p.shape == (number_of_objects_in_hypothesis, state_dimensionality, state_dimensionality)

            assert diff.shape == (number_of_objects_in_hypothesis, number_of_detections, detection_dimensionality)

            # Compute the number of child hypothesis of this hypothesis
            # We could also change this
            mhk = int(round(max(1, self.max_number_of_hypothesis * np.exp(self.log_weights[h]))))

            # Compute mhk best sub-assignments of this hypothesis
            to0 = time.time()
            optimal_assignments: List[Tuple[np.ndarray, np.ndarray]] \
                = Optimizer.optimize(cost_matrix=cost_matrix, number_of_elements=mhk)

            to = time.time()
            self.__visualize_assignments(optimal_assignments, detections=z, hypothesis_identifier=h)

            for decent_assignment in optimal_assignments:
                log_weight: np.ndarray = self.log_weights[h]

                dim_new = decent_assignment[0].shape[0]
                new_r.append(np.empty(dim_new, dtype=np.float))
                new_lbl.append(np.empty((dim_new, 2)))
                new_means.append(np.empty([dim_new, state_dimensionality], dtype=np.float))
                new_covs.append(np.empty([dim_new, state_dimensionality, state_dimensionality], dtype=np.float))

                for idx, (track_index, detection_index) in enumerate(np.array(decent_assignment).T):

                    detected = detection_index < len(z)

                    assert track_index < len(self.lbl[h])
                    assert idx < len(new_lbl[-1])
                    assert h < len(self.lbl)

                    new_lbl[-1][idx] = self.lbl[h][track_index]

                    detection_confidence: Optional[float] = None
                    if detected:
                        detection_confidence = confidences[detection_index] #XXX


                    if detected and (self.__get_object_age(self.lbl[h][track_index]) > 1 or detection_confidence > 0.1):

                        # r := 1 (given this hypothesis, the probability of existence of this detected object is 1)
                        new_r[-1][idx] = 1.0

                        # For innovation covariance S := H cov H^T + R
                        # mean := mean + cov H^T S^-1 (z - H mean)
                        #       = mean +   mat_k    (z -   zh  )
                        new_means[-1][idx] = self.mean[h][track_index] + np.dot(mat_k[track_index],
                                                                                diff[track_index, detection_index])

                        # XXX: experimental
                        if detection_confidence < 0.15:
                            lam = detection_confidence / 0.15 * 0.25 + 0.75
                            new_means[-1][idx] = new_means[-1][idx] * lam + (1 - lam) * self.mean[h][track_index]


                        # cov := (cov - cov H^T S^-1) H cov
                        #      = (cov - cov  mat_k  ) H cov
                        #      =           mat_p
                        new_covs[-1][idx] = mat_p[track_index]

                        # Update the log weight of the assignment, that is a part of the updated log weight of the
                        # hypothesis (after normalization)
                        log_weight_assignment: float = -cost_matrix[track_index][detection_index]


                    else:

                        new_means[-1][idx] = self.mean[h][track_index]
                        new_covs[-1][idx] = self.cov[h][track_index]

                        probability_of_existence = self.r[h][track_index]
                        denuminator = 1 - probability_of_existence + probability_of_existence * (1 - self.update.p_d)
                        new_r[-1][idx] = (probability_of_existence * (1 - self.update.p_d)) / denuminator

                        # Tweak: remove the object quicker (XXX  this might have to go to the cost matrix somehow )
                        _, c, prc = self.__count_object(self.mean[h][track_index], self.lbl[h][track_index])
                        thresh = 0.2
                        if c and (prc > (1 - thresh) or prc < thresh):
                            new_r[-1][idx] = 0

                        log_weight_assignment: float = -cost_matrix[track_index][detection_index]

                    if __debug__:

                        if not detected:
                            assert np.abs(np.log(1 - self.r[h][track_index] * self.update.p_d)
                                          - log_weight_assignment) < 1e-9

                        else:
                            mat_h = self.update.transform
                            detection_dimensionality: int = mat_h.shape[0]
                            assert mat_h.shape == (detection_dimensionality, state_dimensionality)

                            # XXX: mod missing
                            # log_weight = log(P^D/P^C * MB(z; r, H mu, H cov H^t + R)
                            evaluation: np.ndarray = self.__evaluate_multibernoulli(
                                z[detection_index][np.newaxis],
                                rs=self.r[h][track_index][np.newaxis],
                                means=mat_h.dot(self.mean[h][track_index])[np.newaxis],
                                covariances=(mat_h.dot(self.cov[h][track_index]).dot(mat_h.T)
                                             + self.update.covariance)[np.newaxis],
                                from_state_space=True,
                                log=True)

                            assert evaluation.shape == (1, 1)
                            evaluation: float = evaluation[0, 0]
                            evaluation += np.log(self.update.p_d)

                            clutter_intensity = self.get_clutter_intensity(confidences=confidences[detection_index])
                            evaluation -= np.log(clutter_intensity)
                            #assert np.abs(evaluation - log_weight_assignment) < 1e-9 #XXX

                    log_weight += log_weight_assignment

                new_log_weights[len(new_lbl) - 1] = log_weight
        self.r = new_r
        self.mean = new_means
        self.cov = new_covs
        self.lbl = new_lbl
        self.log_weights = new_log_weights

        self.normalize_log_weights()

        te = time.time()
        index = self.__i % self.__to.shape[0]
        self.__to[index] = (tc - ts, tc2 - tc, to - to0, te - to)
        self.__i += 1
        if self.__i % 100 == 0:
            arr = self.__to[:self.__i + 1]
            combined = list(zip(('start', 'cost', 'optimize', 'end'), np.average(arr, axis=0)))
            logger.info(', '.join([f'{t}:{f:0.3f}' for t, f in combined]))

    def normalize_log_weights(self) -> None:
        """
        Normalize the log weights such that the sum of the weights equals 1 in a numerically stable version.
        """
        # Numerically stable way to normalize the weights to a sum of 1: (w - max_w) / (log(sum(exp(w - max_w))))
        max_log_weight: float = np.max(self.log_weights[:len(self.lbl)], initial=0)
        # Both operations cannot be combined, because the updated self.log_weights are utilized in the second operation.
        self.log_weights[:len(self.lbl)] -= max_log_weight
        self.log_weights[:len(self.lbl)] -= np.log(np.sum(np.exp(self.log_weights[:len(self.lbl)])))

        # np.exp(ll - max_log_weight)

        if __debug__:
            total_sum = np.sum(np.exp(self.log_weights[:len(self.lbl)]), initial=1)
            #assert np.abs(total_sum -1) < 1e-9 # XXX: this is wrong :(

    def __compute_cost_matrix_for_hypothesis(self, detections,
                                             hypothesis_identifier: int, flat: bool = False,
                                             confidences: Optional[np.ndarray] = None) \
            -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        return self.__compute_cost_matrix(detections, existence_probabilities=self.r[hypothesis_identifier],
                                          means=self.mean[hypothesis_identifier],
                                          covariance_matrices=self.cov[hypothesis_identifier], flat=flat,
                                          confidences=confidences)

    def get_clutter_intensity(self, confidences: Optional[np.ndarray]) -> Union[float, np.ndarray]:
        if confidences is None:
            return self.update.p_c

        #return self.update.p_c
        return np.where(confidences < 1. / 3, self.update.p_c * (1 + 5 * (1. / 3 - confidences)), self.update.p_c)

    def __compute_cost_matrix(self, detections, *, existence_probabilities: np.ndarray, means: np.ndarray,
                              covariance_matrices: np.ndarray, flat: bool,
                              confidences: Optional[np.ndarray] = None) \
            -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]]:

        number_of_objects: int = len(means)
        number_of_detections: int = len(detections)
        state_dimensionality: int = means.shape[1]
        assert means.shape == (number_of_objects, state_dimensionality)
        assert covariance_matrices.shape == (number_of_objects, state_dimensionality, state_dimensionality)
        assert existence_probabilities.shape == (number_of_objects,)

        assert confidences is None or confidences.shape == (number_of_detections,)

        # Define the cost matrix
        nll_misdetection = -np.log(1 - self.update.p_d * existence_probabilities)
        if flat:
            cost_matrix = np.empty([number_of_objects, number_of_detections + 1])
            cost_matrix[:, -1] = nll_misdetection

        else:
            # Initialize the area of the matrix that models misdetections.
            # weight_misdetection = 1 - p_d * r_i    if i == j
            #                     = np.inf           else
            cost_matrix = np.empty([number_of_objects, number_of_detections + number_of_objects])
            cost_matrix[:, number_of_detections:] = np.inf
            np.fill_diagonal(cost_matrix[:, number_of_detections:], nll_misdetection)

        # Conduct computations for the detection weights
        mat_h = self.update.transform
        detection_dimensionality: int = mat_h.shape[0]
        assert mat_h.shape == (detection_dimensionality, state_dimensionality)

        mat_r = self.update.covariance
        assert mat_r.shape == (detection_dimensionality, detection_dimensionality)

        # means projected into the room of measurements
        # zh = H mu
        zh = np.einsum('ds,...s-> ...d', mat_h, means)
        assert zh.shape == (number_of_objects, detection_dimensionality)
        assert all([1e-10 > np.linalg.norm(zh[i] - np.dot(mat_h, mean)) for i, mean in enumerate(means)])

        # mat_ss = H cov H.T + R
        mat_ss = np.einsum('ab,...bc,ec->...ae', mat_h, covariance_matrices, mat_h) + mat_r
        assert mat_ss.shape == (number_of_objects, detection_dimensionality, detection_dimensionality)

        # __def = mat_h.dot(cov).dot(mat_h.T) + mat_r
        if __debug__:
            __debug_mat_ss = np.array([mat_h.dot(cov).dot(mat_h.T) + mat_r for cov in covariance_matrices]).reshape(
                (number_of_objects, detection_dimensionality, detection_dimensionality))
            __diff = np.abs(__debug_mat_ss - mat_ss)
            __visualization = np.hstack((__debug_mat_ss.reshape((-1, __debug_mat_ss.shape[-1])),
                                         mat_ss.reshape((-1, mat_ss.shape[-1]))))
            assert np.max(np.abs(__diff), initial=0) < 1e-9

        # S^-1
        ssi = np.linalg.pinv(mat_ss)
        assert ssi.shape == (number_of_objects, detection_dimensionality, detection_dimensionality)

        #  N(z; Hmu ; S)
        clutter_intensity: Union[float, np.ndarray] = self.get_clutter_intensity(confidences)
        assert isinstance(clutter_intensity, float) or clutter_intensity.shape == (number_of_detections,)

        # Cost_matrix = -log(r * P^D /c N( z; H mu; S (= HPH^T + R))
        #             = -log(r) - log(pd) + log(c) * logN( z; H mu; S (= HPH^T + R))
        dimension_modulo = None
        if self.detection_orientation_column is not None \
                and self.detection_orientation_column <= detection_dimensionality:
            dimension_modulo = [(self.detection_orientation_column, np.pi)]


        cost_matrix[:, :number_of_detections], ssi, difference_detection_mean = \
            log_evaluate_gaussians_2(x=detections, means=zh, covariance_matrices=mat_ss,
                                       dimensions_modulo=dimension_modulo)


        if __debug__:
            a, b, _, c = log_evaluate_gaussians( detections, zh, mat_ss, return_components=True,
                                                 dimensions_modulo=dimension_modulo)

            __diff__ = a - cost_matrix[:, :number_of_detections]
            assert np.all(np.abs(__diff__) < 1e-9)

            __diff__ = b - ssi
            assert np.all(np.abs(__diff__) < 1e-9)

            __diff__ = c - difference_detection_mean
            assert np.all(np.abs(__diff__) < 1e-9)

        cost_matrix[:, :number_of_detections] = -(cost_matrix[:, :number_of_detections]
                                                  + np.log(existence_probabilities)[:, np.newaxis]
                                                  + np.log(self.update.p_d) - np.log(clutter_intensity))

        if flat:
            return cost_matrix

        if __debug__:
            # cost matrix computed in an expensive way to double check that everything is correct.
            # This cost matrix is positive, hence it has to approximately equal "-cost_matrix[:,:number_of_detections]".
            alternative_costs = np.zeros((number_of_objects, number_of_detections))
            for i in range(number_of_objects):
                mssi = mat_h.dot(covariance_matrices[i]).dot(mat_h.T) + mat_r
                mssii = np.linalg.inv(mssi)

                for j in range(number_of_detections):
                    fu_diff = difference_detection_mean[i][j]

                    value = np.log(existence_probabilities[i]) + np.log(self.update.p_d) - np.log(clutter_intensity[j])
                    value -= 0.5 * np.log((2 * np.pi) ** detection_dimensionality * np.linalg.det(mssi))
                    value -= 0.5 * (fu_diff).dot(mssii).dot(fu_diff.T)

                    alternative_costs[i, j] = value

            difference = alternative_costs + cost_matrix[:, :number_of_detections]
            scalar_difference = np.max(np.abs(difference), initial=0)
            assert scalar_difference < 1e-5

        #
        # Compute helper values for setting updated covariance_matrices and means to avoid duplicate computations.
        #

        # mean := mean + cov H^T S^-1 (z - H mean)
        #       = mean +   K    (z -   zh  )

        # K = cov H^T S^-1
        mat_k = np.einsum('nab,cb,ncd->nad', covariance_matrices, mat_h, ssi)
        assert mat_k.shape == (number_of_objects, state_dimensionality, detection_dimensionality)
        if __debug__:
            __debug_mat_k = np.array(
                [covariance_matrices[i].dot(mat_h.T).dot(ssi[i]) for i in range(number_of_objects)]) \
                .reshape((number_of_objects, state_dimensionality, detection_dimensionality))
            __diff = np.abs(__debug_mat_k - mat_k)
            assert np.max(__diff, initial=0) < 1e-9

        # cov := cov - (cov H^T S^-1) H cov
        #      = cov - (    mat_k   ) H cov
        mat_p = np.einsum('nad,ds,nsS->naS', mat_k, mat_h, covariance_matrices)
        assert mat_p.shape == (number_of_objects, state_dimensionality, state_dimensionality)
        mat_p = covariance_matrices - mat_p

        if __debug__:
            __debug_mat_p = \
                np.array([covariance_matrices[i] - covariance_matrices[i].dot(mat_h.T).dot(ssi[i]).dot(mat_h).dot(
                    covariance_matrices[i])
                          for i in range(number_of_objects)]).reshape((
                    number_of_objects, state_dimensionality, state_dimensionality
                ))

            __diff = np.abs(__debug_mat_p - mat_p)
            assert np.max(__diff, initial=0) < 1e-9

        return cost_matrix, zh, mat_k, mat_p, difference_detection_mean

    def __reduce(self) -> List[np.ndarray]:

        # Capping
        arg_sorted: List[int] = sorted(range(len(self.lbl)), key=self.log_weights[:len(self.lbl)].__getitem__,
                                       reverse=True)

        dead_labels: List[np.ndarray] = []

        # Pruning
        i: int = 0
        for i, lw in enumerate(arg_sorted[::-1]):
            if lw >= self.prune_log_weight or i >= self.max_number_of_hypothesis:
                break
        if i:

            # release colors
            for hypothesis_index in arg_sorted[-i:]:
                labels = self.lbl[hypothesis_index]
                dead_labels.extend(labels)

            arg_sorted = arg_sorted[:-i]

        self.r = [self.r[i] for i in arg_sorted]
        self.mean = [self.mean[i] for i in arg_sorted]
        self.cov = [self.cov[i] for i in arg_sorted]
        self.lbl = [self.lbl[i] for i in arg_sorted]

        # Bernoulli pruning
        for h in self.active_hypothesis_ids():

            ri: int = 0
            while ri < len(self.r[h]):
                r: float = self.r[h][ri]

                # remove all objects that have got a probability of existence lower than the configured prune
                # probability.
                object_pruned = r < self.object_prune_probability

                # Remove all objects that stem from a birth without detections as parents
                object_age: int = self.__get_object_age(self.lbl[h][ri])
                object_pruned = object_pruned or ((r != 1) and (object_age == 0))

                # _dir, _count, prc = self.__count_object(self.mean[h][ri], self.lbl[h][ri])
                # object_pruned = object_pruned or (prc > 0.9 and r < 1)

                if object_pruned:
                    dead_labels.append(self.lbl[h][ri])
                    self.r[h] = np.delete(self.r[h], ri, axis=0)
                    self.mean[h] = np.delete(self.mean[h], ri, axis=0)
                    self.cov[h] = np.delete(self.cov[h], ri, axis=0)
                    self.lbl[h] = np.delete(self.lbl[h], ri, axis=0)
                else:
                    ri += 1

        self.normalize_log_weights()
        return dead_labels

    def __get_object_age(self, object_label: Union[np.ndarray, Tuple[int, int]]) -> int:
        return int(self.time_count - object_label[0])

    def __estimate(self) -> Dict[Tuple[int, int], Tuple[np.ndarray, float, np.ndarray]]:

        estimates: Dict[Tuple[int, int], Tuple[np.ndarray, float, np.ndarray]] = dict()

        if len(self.lbl):

            # called after weight normalization -> best hypothesis = hypothesis 0
            best_hypothesis: int = 0

            for i, f in enumerate(self.r[best_hypothesis]):
                if f > self.object_exists_probability:

                    object_age: int = self.__get_object_age(self.lbl[best_hypothesis][i])
                    if object_age >= 1:
                        estimates[tuple(self.lbl[best_hypothesis][i])] = (self.mean[best_hypothesis][i],
                                                                          self.r[best_hypothesis][i],
                                                                          self.cov[best_hypothesis][i])

        return estimates

    def __count_object(self, position, identifier: Tuple[int, int]) -> Tuple[int, int, float]:

        count_id = identifier[1]
        assert count_id < self.predict.birth.get_number_of_birth_components()
        top_component = count_id < self.predict.birth.get_number_of_birth_components() / 2

        id_position_offset = 1 if self.horizontal else 0
        fov = self.predict.transition.survival_probability.fov
        v1, v2 = fov[id_position_offset], fov[id_position_offset + 2]
        s = v2 - v1

        count = position[id_position_offset] > s / 2 if top_component else position[id_position_offset] < s / 2
        prc = position[id_position_offset] / s if top_component else (s - position[id_position_offset]) / s
        outgoing = top_component

        return outgoing, int(count), prc

    def __get_track_color(self, tid: Tuple[int, int], grey_birth: bool = True) -> Tuple[int, int, int]:
        assert len(tid) == 2

        if tid not in self.__used_colors:
            if len(self.__colors):
                self.__used_colors[tid] = self.__colors.pop()
            else:
                self.__used_colors[tid] = DISTINCT_TRACKING_COLORS[hash(tuple(tid)) % len(DISTINCT_TRACKING_COLORS)]

        # For birth bernoullis don't swap colors
        if grey_birth and tid[0] + 1 >= self.time_count:
            return (125, 125, 125)

        assert len(self.__used_colors[tid]) == 3
        return self.__used_colors[tid]

    def __count(self, new_estimates: Dict[Tuple[int, int], Tuple[np.ndarray, float, np.ndarray]]) \
            -> Tuple[int, int, List[Tuple[int, int]]]:
        updated_tracks: Dict[Tuple[int, int], Tuple[np.ndarray, float, Tuple[int, int, int], int]] = dict()

        eid: Tuple[int, int]
        estimate: Tuple[np.ndarray, float, np.ndarray]
        for eid, estimate in new_estimates.items():
            last_mean, probability_of_existence, _cov = estimate
            currently_observed: int = 1 if probability_of_existence == 1.0 else 0

            if eid in self.tracks:
                positions, _old_probability_of_existence, color, observe_count = self.tracks.pop(eid)
                positions = np.vstack((positions, last_mean))

                number_of_positions, _state_dimensionality = positions.shape
                if number_of_positions > 200:
                    positions = positions[::2].copy()

                updated_tracks[eid] = (positions, probability_of_existence, color, observe_count + currently_observed)

            else:
                updated_tracks[eid] = (last_mean[np.newaxis], probability_of_existence, self.__get_track_color(eid, False),
                                       currently_observed)

        # dead estimates
        new_outgoing: int = 0
        new_incoming: int = 0
        numbers_of_observations_and_ages: List[Tuple[int, int]] = []
        for eid, (pos, mu, clr, number_of_observations) in self.tracks.items():

            # XXX: min age requirement
            # XXX: min number of detection requirement (...)

            outgoing, count, _prc = self.__count_object(pos[-1], eid)

            if count != 0:
                numbers_of_observations_and_ages.append((number_of_observations, self.__get_object_age(eid)))

            if number_of_observations > 15:
                if outgoing:
                    new_outgoing += count
                else:
                    new_incoming += count
            """
            start = pos[0]
            stop = pos[-1]

            id_offset = 1 if self.horizontal else 0

            fov = self.predict.transition.survival_probability.fov

            x, x2 = fov[id_offset], fov[id_offset + 1]
            w = x2 - x

            if start[id_offset] < w / 2 < stop[id_offset]:
                new_outgoing += 1
            elif start[id_offset] > w / 2 > stop[id_offset]:
                new_incoming += 1
            """

        # XXX: when there are multiple hypothesis, here we need to prune the dead estimates aswell from other hypothesis.

        self.tracks = updated_tracks

        self.count_in += new_incoming
        self.count_out += new_outgoing

        return new_incoming, new_outgoing, numbers_of_observations_and_ages

    def active_hypothesis_ids(self):
        return range(len(self.lbl))

    def __call__(self, xs, *, hypothesis_identifier: int = 0, from_state_space: bool = False) -> np.ndarray:
        """
        Evaluate `xs` at the distribution of a single or all hypothesis and marginalize along different objects.

        :param xs:                     numpy array of all the values at which the distribution is to be evaluated.
        :param hypothesis_identifier:  identifier of the hypothesis that is to be evaluated or -1 to visualize all
                                       hypothesis in a joined density.
        :param from_state_space:       True iff the xs are samples from the space of the distributions (if False, the
                                       distributions have to undergo transformations)
        :return:                       the joint probability density evaluated at xs of shape (number_of_xs, ).
        """
        hypothesis_to_be_visualized: List[int] = list(self.active_hypothesis_ids()) if hypothesis_identifier == -1 \
            else [hypothesis_identifier] if hypothesis_identifier < len(self.lbl) else []

        number_of_values = xs.shape[0]
        values = np.zeros(number_of_values)
        weights = np.exp(self.log_weights[:len(self.lbl)])
        for h in hypothesis_to_be_visualized:
            evaluated_hypothesis = self.__evaluate_hypothesis(xs=xs, hypothesis_identifier=h,
                                                              from_state_space=from_state_space)
            evaluated_hypothesis = evaluated_hypothesis.sum(axis=0)
            assert evaluated_hypothesis.shape == (number_of_values,)
            values += weights[h] * evaluated_hypothesis

        assert values.shape == (number_of_values,)
        return values

    def __log_evaluate_hypothesis(self, xs: np.ndarray, hypothesis_identifier: int,
                                  from_state_space: bool = False) -> np.ndarray:
        """
        Compute the log value the normal distributions of the hypothesis identified with `hypothesis_identifier` assume
        at the vales `xs`

        :param xs:                    numpy array of all the values at which the hypothesis is to be evaluated
        :param hypothesis_identifier: the identifier of the hypothesis to be evaluated
        :param from_state_space:      True iff the xs are samples from the space of the distributions (if False, the
                                      distributions have to undergo transformations)
        :return:                      numpy array of shape (number_of_distributions_in_hypothesis, number_of_xs)
        """
        return self.__evaluate_multibernoulli(xs=xs, rs=self.r[hypothesis_identifier],
                                              means=self.mean[hypothesis_identifier],
                                              covariances=self.cov[hypothesis_identifier],
                                              from_state_space=from_state_space, log=True)

    def __evaluate_hypothesis(self, xs: np.ndarray, hypothesis_identifier: int,
                              from_state_space: bool = False) -> np.ndarray:
        """
        Compute the value the normal distributions of the hypothesis identified with `hypothesis_identifier` assume
        at the vales `xs`

        :param xs:                    numpy array of all the values at which the hypothesis is to be evaluated
        :param hypothesis_identifier: the identifier of the hypothesis to be evaluated
        :param from_state_space:      True iff the xs are samples from the space of the distributions (if False, the
                                      distributions have to undergo transformations)
        :return:                      numpy array of shape (number_of_distributions_in_hypothesis, number_of_xs)
        """
        return self.__evaluate_multibernoulli(xs=xs, rs=self.r[hypothesis_identifier],
                                              means=self.mean[hypothesis_identifier],
                                              covariances=self.cov[hypothesis_identifier],
                                              from_state_space=from_state_space, log=False)

    def __evaluate_multibernoulli(self, xs: np.ndarray, rs: np.ndarray, means: np.ndarray, covariances: np.ndarray,
                                  from_state_space: bool = False, log: bool = False) -> np.ndarray:
        """
        Compute the value the normal distributions of the hypothesis identified with `hypothesis_identifier` assume
        at the vales `xs`

        :param xs:                    numpy array of all the values at which the hypothesis is to be evaluated
        :param rs:                    The existence probability of the MB.
        :param means:                 The means of the MB.
        :param covariances:           The covariance matrices of the MB.
        :param from_state_space:      True iff the xs are samples from the space of the distributions (if False, the
                                      distributions have to undergo transformations)
        :param log:                   true if the return value should be the logarithm of the value
        :return:                      numpy array of shape (number_of_distributions_in_hypothesis, number_of_xs)
        """
        number_of_distributions: int = means.shape[0]
        state_dimensionality: int = means.shape[1]
        assert rs.shape == (number_of_distributions,)
        assert means.shape == (number_of_distributions, state_dimensionality)
        assert covariances.shape == (number_of_distributions, state_dimensionality, state_dimensionality)

        number_of_values: int = xs.shape[0]

        if from_state_space:
            assert xs.shape[0] == number_of_values
            assert xs.shape[1] <= state_dimensionality

        else:
            mat_h = self.update.transform
            detection_dimensionality: int = mat_h.shape[0]
            assert mat_h.shape == (detection_dimensionality, state_dimensionality)
            assert xs.shape[0] == number_of_values
            assert xs.shape[1] <= detection_dimensionality

            means = np.einsum('ds,ns->nd', mat_h, means)
            assert means.shape == (number_of_distributions, detection_dimensionality)

            covariances = np.einsum('ds,nss,Ds->ndD', mat_h, covariances, mat_h) + self.update.covariance
            assert covariances.shape == (number_of_distributions, detection_dimensionality, detection_dimensionality)

        evaluation_function = log_evaluate_gaussians if log else evaluate_gaussians

        values = evaluation_function(xs, means=means, covariance_matrices=covariances)
        assert values.shape == (number_of_distributions, number_of_values)

        values = (values + np.log(rs[:, np.newaxis])) if log else (values * rs[:, np.newaxis])
        assert values.shape == (number_of_distributions, number_of_values)

        return values

    def __visualize_assignments(self, optimal_assignments, detections, hypothesis_identifier) -> None:
        """
        Visualize the detections as green circles and the object positions as markers. The x  is black if detectesd, red
        otherwise.

        If an object is associated to a measurement in this iteration, draw a line between the measurement and the
        object.
        """
        if self.plot:

            # Utilize the same window
            first = not isinstance(self.plot, tuple)
            if first:
                self.plot = plt.figure()
                self.plot = (self.plot, self.plot.add_subplot(1, 1, 1))

            fig, ax = self.plot
            if not first:
                ax.clear()

            width = self.predict.transition.survival_probability.fov[3]
            zs = detections.copy()
            zs[:, 1] = width - zs[:, 1]

            # Plot measurements as green circles
            ax.scatter(*zs.T, color='green', s=250., alpha=.3)

            for track_index, detection_index in np.array(optimal_assignments[0]).T:
                position = self.mean[hypothesis_identifier][track_index]
                tid = self.lbl[hypothesis_identifier][track_index]
                detected = detection_index < len(detections)

                if detected:
                    detection = detections[detection_index] if detected else None

                    ax.plot([position[0], detection[0]], [width - position[1], width - detection[1]],
                            color=get_track_color(tid, as_float=True))

                    ax.scatter([position[0], detection[0]], [width - position[1], width - detection[1]],
                               color=get_track_color(tid, as_float=True), marker='x')

                elif track_index >= self.predict.birth.get_number_of_birth_components():
                    ax.scatter(position[0], width - position[1], color='red', marker='o')

            # draw the boundaries of the field of view via two points.
            ax.scatter(np.array(self.predict.transition.survival_probability.fov)[[0, 2]],
                       np.array(self.predict.transition.survival_probability.fov)[[1, 3]])
            if first:
                fig.show()
            fig.canvas.draw()

            l = ... == ...
            # plt.pause(0.05)
            # plt.close()
            # plt.show(block=False)

    def get_state_dimensionality(self) -> int:
        _detection_dimensionality, state_dimensionality = self.update.transform.shape
        return state_dimensionality

    def get_detection_dimensionality(self) -> int:
        detection_dimensionality, _state_dimensionality = self.update.transform.shape
        return detection_dimensionality

    def __evaluate_heat(self, image_size: Tuple[int, int],
                        original_image_size: Optional[Tuple[int, int]] = None) -> np.ndarray:
        if original_image_size is None:
            original_image_size = image_size

        grid = np.moveaxis(np.array(np.meshgrid(np.linspace(1, original_image_size[1], image_size[1]),
                                                np.linspace(0, original_image_size[0], image_size[0]))), 0, 2)
        # grid = np.pad(grid, [(0, 0), (0, 0), (0, self.get_detection_dimensionality() - 2)])
        # assert grid.shape[-1] == self.get_detection_dimensionality()

        original_grid_shape = grid.shape
        grid = grid.reshape(-1, 2)  # self.get_detection_dimensionality())
        number_of_grid_points: int = grid.shape[0]

        values = self.__call__(grid, hypothesis_identifier=0, from_state_space=False)
        # values = self.__log_evaluate_hypothesis(xs=grid, hypothesis_identifier = 0, from_state_space=False)
        # values = values.sum(axis=0)

        values = values.reshape(original_grid_shape[:2])

        min_val = np.min(values)
        max_val = np.max(values)
        if max_val != min_val:
            np.subtract(values, min_val, out=values)
            np.divide(values, max_val - min_val, out=values)
        cmap = plt.get_cmap('jet')
        image = (np.delete(cmap(values), 3, 2) * 255).astype(np.uint8)
        return image

        return values

    def __evaluate_argmax(self, labels: np.ndarray, existence_probabilities: np.ndarray, means: np.ndarray,
                          covariance_matrices: np.ndarray,
                          image_size: Tuple[int, int],
                          original_image_size: Optional[Tuple[int, int]] = None) -> np.ndarray:
        if original_image_size is None:
            original_image_size = image_size

        grid = np.moveaxis(np.array(np.meshgrid(np.linspace(1, original_image_size[1], image_size[1]),
                                                np.linspace(0, original_image_size[0], image_size[0]))), 0, 2)
        original_grid_shape = grid.shape
        grid = grid.reshape(-1, 2)
        number_of_grid_points: int = grid.shape[0]

        nll_assignments = self.__compute_cost_matrix(grid, existence_probabilities=existence_probabilities,
                                                     means=means, covariance_matrices=covariance_matrices, flat=True)
        number_of_distributions: int = nll_assignments.shape[0]
        assert nll_assignments.shape == (number_of_distributions, number_of_grid_points + 1)

        # remove the assignments that are worse than misdetections
        nll_misdetection = nll_assignments[:, -1]
        assert nll_misdetection.shape == (number_of_distributions,)
        # nll_assignments = nll_assignments[:, :-1]
        nll_assignments = np.where(nll_assignments[:, :-1] > nll_misdetection[:, np.newaxis], np.inf,
                                   nll_assignments[:, :-1])
        assert nll_assignments.shape == (number_of_distributions, number_of_grid_points)

        best_assignments = np.argmin(nll_assignments, axis=0) if nll_assignments.shape[0] \
            else (np.ones((number_of_grid_points), dtype=np.int) * -1)
        assert best_assignments.shape == (number_of_grid_points,)

        if not number_of_distributions:
            best_assignment_values = np.ones(number_of_grid_points) * -1
        else:
            best_assignment_values = nll_assignments.T[list(range(nll_assignments.shape[1])), best_assignments]
            # try:
            #    best_assignment_values = np.choose(best_assignments, nll_assignments)
            # except Exception as e:
            #    print(e)
            # best_assignment_values = nll_assignments[range(len(best_assignments)), best_assignments]

        assert best_assignment_values.shape == (number_of_grid_points,)

        color_indices = np.where(best_assignment_values != np.inf, best_assignments, -1)
        assert color_indices.shape == (number_of_grid_points,)

        # XXX Color identifiers to rgb colors
        color_lookup = np.array([self.__get_track_color(tuple(tid.astype(np.int))) for tid in labels] + [(0, 0, 0)],
                                dtype=np.uint8)
        colors = color_lookup[color_indices]
        assert colors.shape == (number_of_grid_points, 3)
        colors = colors.reshape((*original_grid_shape[:2], 3))

        return colors

    def __visualize_trajectories(self, image, *,
                                 estimates: Optional[np.ndarray] = None,
                                 all_detections: Optional[Dict[Tuple[int, int],
                                                               Tuple[np.ndarray, float, np.ndarray]]] = None,
                                 tracks: Optional[Dict[Tuple[int, int],
                                                       Tuple[np.ndarray, float, Tuple[int, int, int]]]] = None,
                                 evaluate_fully: bool = False) -> None:
        """

        :param image:
        :param all_detections:  All detections that are theoretically available (inlcuding those that are excluded
                                because they are filtered out because of a too low confidence.

        :param estimates:       All produced estimates

        :param tracks:          All currently active tracks

        :param evaluate_fully:  True iff the image that is to be produced should consist of a probability density
                                map.
        """

        # Fetch underground
        if evaluate_fully:
            image_height, image_width = image.shape[:2]
            resolution = 1.0
            grid = np.mgrid[0: image_width: resolution, 0: image_height: resolution].T
            w, h = grid.shape[0], grid.shape[1]
            grid = grid.reshape([w * h, grid.shape[2]]), w, h
            values = self.evaluate(grid, hypothesis_identifier=0)
            image = values.reshape((w, h))

            convert: bool = True
            if convert:
                mn, mx = np.min(image), np.max(image)
                image -= mn
                if mx - mn:
                    image /= (mx - mn)

                cmap = plt.get_cmap('jet')
                image = (np.delete(cmap(image), 3, 2) * 255).astype(np.uint8)

        else:
            pmf = image.copy()

        if all_detections is not None:
            self.__visualize_detections(image, detections=all_detections)

        if tracks is not None:
            self.__visualize_tracks(image, tracks=tracks)

        if estimates is not None:
            self.__visualize_estimates(image, estimates=estimates)

        if tracks is not None:
            ...

        return pmf

    def __visualize_estimates(self, image: np.ndarray,
                              estimates: Dict[Tuple[int, int], Tuple[np.ndarray, float, np.ndarray]],
                              visualize_covariance_matrix: bool = False,
                              original_image_size: Optional[Tuple[int, int]] = None):

        image_size = image.shape[:2]
        scale_factor: np.ndarray = np.ones(2) if original_image_size is None or image_size is None \
            else np.array(image.shape[:2], dtype=np.float32) / np.array(original_image_size, dtype=np.float32)

        tid: Tuple[int, int]
        marker: Tuple[np.ndarray, float, np.ndarray]
        for tid, marker in estimates.items():

            try:
                color: Tuple[int, int, int] = self.__get_track_color(tid)
                mean = marker[0]

                if visualize_covariance_matrix:
                    covariance = marker[-1]

                    draw_covariance_ellises(image, means=mean[np.newaxis],
                                            covariance_matrices=covariance[np.newaxis],
                                            probabilities=[0.8, 0.9, 0.99], colors=[color], rs=None,
                                            original_image_size=original_image_size, image_size=image_size)

                mean = (marker[0][:2] * scale_factor).astype(np.int)
                cv2.drawMarker(image, tuple(mean[:2]), (100, 100, 100), cv2.MARKER_STAR,
                               int(max(1, 15 * scale_factor.min(initial=1))), 1)
            except AssertionError as e:
                raise e from None

            except Exception as e:
                raise e from None
                logger.warning(f'Error drawing estimate. Position outside the image? {e}')

    def __visualize_states(self, image: np.ndarray, *, labels: np.ndarray, means: np.ndarray,
                           covariance_matrices: np.ndarray, existence_probabilities: np.ndarray,
                           visualize_covariance_matrix: bool = False, visualize_existence_probabilities: bool = False,
                           visualize_state_size: bool = False, visualize_label: bool = True,
                           original_image_size: Optional[Tuple[int, int]] = None):

        number_of_distributions: int = labels.shape[0]
        state_dimensionality: int = means.shape[1]
        assert existence_probabilities.shape == (number_of_distributions,)
        assert means.shape == (number_of_distributions, state_dimensionality)
        assert covariance_matrices.shape == (number_of_distributions, state_dimensionality, state_dimensionality)

        image_size = image.shape[:2]
        scale_factor: np.ndarray = np.ones(2) if original_image_size is None \
            else np.array(image.shape[:2], dtype=np.float32) / np.array(original_image_size, dtype=np.float32)

        for i_d, tid in enumerate(labels):

            try:
                color: Tuple[int, int, int] = self.__get_track_color(tuple(tid.astype(np.int)))

                is_birth_bernoulli = tid[0] + 1 >= self.time_count
                if is_birth_bernoulli:
                    continue

                if visualize_covariance_matrix:
                    covariance_matrix = covariance_matrices[i_d]

                    draw_covariance_ellises(image, means=means[i_d][np.newaxis],
                                            covariance_matrices=covariance_matrix[np.newaxis],
                                            probabilities=[0.8, 0.9, 0.99], colors=[color], rs=None,
                                            image_size=image_size, original_image_size=original_image_size)

                mean = (means[i_d][:2] * scale_factor).astype(np.int)
                cv2.drawMarker(image, tuple(mean), (100, 100, 100), cv2.MARKER_STAR,
                               int(max(1, 15 * scale_factor.min(initial=1))), 1)
                if self.state_size_columns is not None and self.state_orientation_column is not None \
                        and visualize_state_size:
                    state_size: np.array = means[i_d][list(self.state_size_columns)]
                    orientation = means[i_d][self.state_orientation_column][np.newaxis]

                    top = state_size * [self.detection_size_divided_by / 2, 0]
                    right = state_size * [0, self.detection_size_divided_by / 2]
                    top_right = np.vstack((top, right))[np.newaxis]
                    assert top_right.shape == (1, 2, 2)

                    rotate(top_right, -orientation)
                    assert top_right.shape == (1, 2, 2)
                    top, right = (mean + top_right[0]).astype(np.int)

                    cv2.line(image, tuple(mean), tuple(top), (255, 255, 255), 1, cv2.LINE_4)
                    cv2.line(image, tuple(mean), tuple(right), (0, 0, 0), 1, cv2.LINE_4)

                if visualize_existence_probabilities or visualize_label:
                    text = ''
                    if visualize_label:
                        text += str(tid)
                    if visualize_existence_probabilities:
                        text += f'{existence_probabilities[i_d]:.2f}'

                    cv2.putText(image, text, tuple(mean), cv2.FONT_HERSHEY_SIMPLEX, 0.6 *
                                scale_factor.min(initial=1), (0, 0, 0), 1, cv2.LINE_8)

            except AssertionError as e:
                raise e from None

            except Exception as e:
                raise e from None
                logger.warning(f'Error drawing estimate. Position outside the image? {e}')

    def __visualize_detections(self, image: np.ndarray,
                               detections: np.ndarray,
                               original_image_size: Optional[Tuple[int, int]] = None,
                               draw_detection_size: bool = False) -> None:
        """
        Draw 'x' marker for each detection, colored in red, orange, yellow and black if confidence is contained
        in the detections to give visual feedback on the detection's confidence.

        :param image:                   RGB image.
        :param detections:              all detections that are to be visualized (may be more than given to the
                                        tracking algorithm).
        :param original_image_size:     Size of the original image from which we took the detections in case the
                                        detections have to be scaled.
        """
        image_size = np.array(image.shape[:2], dtype=np.float32)
        scale_factor: np.ndarray = np.ones(2) if original_image_size is None \
            else image_size / np.array(original_image_size, dtype=np.float32)

        number_of_detections: int = detections.shape[0]
        detection_dimensionality: int = detections.shape[1]
        assert detections.shape == (number_of_detections, detection_dimensionality)

        for detection in detections:
            detection_position: np.ndarray = (detection[:2] * scale_factor).astype(int)
            color: str = (255, 255, 255)

            # Adapt detection color if confidence is attached
            if self.detection_confidence_column is not None:
                detection_confidence: float = detection[self.detection_confidence_column]
                color: str = \
                    (255, 0, 0) if detection_confidence <= 0.1 else \
                        (255, 140, 0) if detection_confidence <= 0.2 else \
                            (255, 215, 0) if detection_confidence <= 0.3 else color

            cv2.drawMarker(image, tuple(detection_position), color, cv2.MARKER_CROSS,
                           int(max(1, 15 * scale_factor.min(initial=1))), 1)
            if self.detection_size_columns is not None and self.detection_orientation_column is not None \
                    and draw_detection_size:
                detection_size: np.array = detection[list(self.detection_size_columns)]
                orientation = detection[self.detection_orientation_column][np.newaxis]

                top = detection_size * [self.detection_size_divided_by / 2, 0]
                right = detection_size * [0, self.detection_size_divided_by / 2]
                top_right = np.vstack((top, right))[np.newaxis]
                assert top_right.shape == (1, 2, 2)

                rotate(top_right, -orientation)
                assert top_right.shape == (1, 2, 2)
                top, right = (detection_position + top_right[0]).astype(np.int)

                cv2.line(image, tuple(detection_position), tuple(top), (255, 255, 255), 1, cv2.LINE_4)
                cv2.line(image, tuple(detection_position), tuple(right), (150, 150, 150), 1, cv2.LINE_4)

    def __visualize_tracks(self, image,
                           tracks: Dict[Tuple[int, int], Tuple[np.ndarray, float, Tuple[int, int, int]]],
                           original_image_size: Optional[Tuple[int, int]] = None) -> np.ndarray:
        scale_factor: np.ndarray = np.ones(2) if original_image_size is None \
            else np.array(image.shape[:2], dtype=np.float32) / np.array(original_image_size, dtype=np.float32)

        track_identifier: Tuple[int, int]
        track: Tuple[np.ndarray, float, Tuple[int, int, int]]

        for i, (track_identifier, track) in enumerate(tracks.items()):
            positions: np.ndarray
            existence_probability: bool
            track: Tuple[int, int, int]

            positions, existence_probability, clr, _object_observation_count = track
            number_of_positions: int = positions.shape[0]
            state_dimensions: int = positions.shape[1]
            assert positions.shape == (number_of_positions, state_dimensions)

            adapted_positions = (positions[:, :2] * scale_factor)[np.newaxis].astype(np.int32, copy=True)
            cv2.polylines(image, list(adapted_positions), isClosed=False, color=clr,
                          thickness=1, lineType=cv2.LINE_8)

            visualize_textual_information: bool = False
            if visualize_textual_information:
                cv2.putText(image, f'{track_identifier} {existence_probability}',
                            tuple(positions[-1][:2]), cv2.FONT_HERSHEY_SIMPLEX, 0.6 * scale_factor.min(initial=1),
                            clr, 1, cv2.LINE_8)

        return image

    def __visualize_tic(self, visualization_image, __states: List, *, all_detections,
                        hypothesis_identifier: int = 0) -> Optional[np.ndarray]:

        # If no visualization image is provided, don't compute anything
        if visualization_image is None:
            return None

        # If no states are provided, only visualize the tracks
        if not len(__states) or not self.__visualize_fully:
            self.__visualize_tracks(visualization_image, self.tracks)
            self.__visualize_detections(visualization_image, detections=all_detections, draw_detection_size=True)
            return visualization_image

        if self.__last_visualization is not None and not len(all_detections) and not len(
                self.mean[hypothesis_identifier]):
            return self.__last_visualization

        # Perform full visualization of the intermediate states.
        assert len(__states) == 4
        prior_states, predict_states, update_states, reduced_states = __states

        original_image_size: Tuple[int, int] = visualization_image.shape[:2]
        reduced_image_size: Tuple[int, int] = tuple((np.array(original_image_size) / 4).astype(np.int))

        interpol = cv2.INTER_NEAREST

        distance_between_images: int = 5
        w, h = original_image_size  # reduced_image_size
        dw, dh = (w + distance_between_images), h + distance_between_images

        v1 = False
        v2 = True
        v3 = False

        number_rows: int = 3
        number_cols: int = v1 + v2 + v3
        adapted_image: np.ndarray = np.ones((dw * number_rows, dh * number_cols, 3), dtype=np.uint8)

        def mat(i_w: int, i_h: int, dest: Optional[np.ndarray] = None, update=lambda a, b: b,
                s_w: int = 1, s_h: int = 1) -> np.ndarray:
            if dest is not None:
                adapted_image[
                i_w * dw: (i_w + s_w) * dw - s_w * distance_between_images,
                i_h * dh: (i_h + s_h) * dh - s_h * distance_between_images] = update(
                    adapted_image[
                    i_w * dw: (i_w + s_w) * dw - s_w * distance_between_images,
                    i_h * dh: (i_h + s_h) * dh - s_h * distance_between_images], dest)

            return adapted_image[
                   i_w * dw: (i_w + s_w) * dw - s_w * distance_between_images,
                   i_h * dh: (i_h + s_h) * dh - s_h * distance_between_images]

        #
        # Visualize the Prior state (means, probability map)
        #
        if v1:
            prior_weight, prior_labels, prior_existence_probabilities, prior_means, prior_covariance_matrices = \
                (ps[hypothesis_identifier] for ps in prior_states) if len(prior_states[1]) else (None,) * 5

            i_w, i_h = 0, 0
            np.copyto(src=visualization_image, dst=mat(i_w, i_h))
            if prior_weight is not None:
                self.__visualize_states(mat(i_w, i_h), labels=prior_labels, means=prior_means,
                                        covariance_matrices=prior_covariance_matrices,
                                        existence_probabilities=prior_existence_probabilities,
                                        visualize_covariance_matrix=False, visualize_existence_probabilities=False,
                                        visualize_state_size=False, original_image_size=original_image_size)
                self.__visualize_tracks(mat(i_w, i_h), tracks=self.tracks, original_image_size=original_image_size)

            i_w, i_h = 1, 0
            if prior_weight is not None:
                prior_argmax = self.__evaluate_argmax(labels=prior_labels,
                                                      existence_probabilities=prior_existence_probabilities,
                                                      means=prior_means, covariance_matrices=prior_covariance_matrices,
                                                      image_size=reduced_image_size,
                                                      original_image_size=original_image_size)

                cv2.resize(prior_argmax, dsize=original_image_size[::-1], dst=mat(i_w, i_h),
                           interpolation=interpol)
                mat(i_w, i_h, visualization_image, lambda a, b: a * 0.6 + b * 0.4)

                self.__visualize_states(mat(i_w, i_h), labels=prior_labels, means=prior_means,
                                        covariance_matrices=prior_covariance_matrices,
                                        existence_probabilities=prior_existence_probabilities,
                                        visualize_covariance_matrix=False, visualize_existence_probabilities=True,
                                        visualize_state_size=True, original_image_size=original_image_size)
        #
        # Visualize the Prediction
        #
        if v2:
            predict_weight, predict_labels, predict_existence_probabilities, predict_means, predict_covariance_matrices = \
                (ps[hypothesis_identifier] for ps in predict_states) if len(predict_states[1]) else (None,) * 5
            """
            i_w, i_h = 0, (1 * v1)
            image: np.ndarray = self.predict.transition.survival_probability.evaluate(
                image_size=reduced_image_size, original_image_size=original_image_size)
            cv2.resize(image[: original_image_size[0], : original_image_size[1]],
                       dsize=original_image_size[::-1], dst=mat(i_w, i_h),
                       interpolation=interpol)
            mat(i_w, i_h, visualization_image, lambda a, b: a * 0.6 + b * 0.4)

            number_birth = self.predict.birth.get_number_of_birth_components()
            draw_covariance_ellises(mat(i_w, i_h), predict_means[number_birth:].copy(),
                                    predict_covariance_matrices[number_birth:],
                                    [0.8, 0.9, 0.99],
                                    list((self.__get_track_color(tuple(a.astype(np.int)))
                                          for a in predict_labels[number_birth:])),
                                    predict_existence_probabilities[number_birth:],
                                    )  # original_image_size=original_image_size, image_size=reduced_image_size)
            """

            i_w, i_h = 1, (1 * v1)
            prediction_argmax = self.__evaluate_argmax(labels=predict_labels,
                                                       existence_probabilities=predict_existence_probabilities,
                                                       means=predict_means.copy(),
                                                       covariance_matrices=predict_covariance_matrices,
                                                       image_size=reduced_image_size,
                                                       original_image_size=original_image_size)
            cv2.resize(prediction_argmax, dsize=original_image_size[::-1], dst=mat(i_w, i_h),
                       interpolation=interpol)
            mat(i_w, i_h, visualization_image, lambda a, b: a * 0.6 + b * 0.4)
            self.__visualize_states(mat(i_w, i_h), labels=predict_labels, means=predict_means,
                                    covariance_matrices=predict_covariance_matrices,
                                    existence_probabilities=predict_existence_probabilities,
                                    visualize_covariance_matrix=False, visualize_existence_probabilities=True,
                                    visualize_state_size=True, original_image_size=original_image_size)
            self.__visualize_detections(mat(i_w, i_h), detections=all_detections, draw_detection_size=False,
                                        original_image_size=original_image_size)

        #
        # Visualize the Update (after reduction)
        #
        if v3:
            update_weight, update_labels, update_existence_probabilities, update_means, update_covariance_matrices = \
                (ps[hypothesis_identifier] for ps in reduced_states) if len(reduced_states[1]) else (None,) * 5

            i_w, i_h = 0, v1 + v2

            heat = self.__evaluate_heat(image_size=reduced_image_size, original_image_size=original_image_size)
            cv2.resize(heat, dsize=original_image_size[::-1], dst=mat(i_w, i_h),
                       interpolation=interpol)

            i_w, i_h = 1, v1 + v2
            if update_labels is not None:
                update_argmax = self.__evaluate_argmax(labels=update_labels,
                                                       existence_probabilities=update_existence_probabilities,
                                                       means=update_means,
                                                       covariance_matrices=update_covariance_matrices,
                                                       image_size=reduced_image_size,
                                                       original_image_size=original_image_size)

                cv2.resize(update_argmax, dsize=original_image_size[::-1], dst=mat(i_w, i_h),
                           interpolation=interpol)

            mat(i_w, i_h, visualization_image, lambda a, b: a * 0.6 + b * 0.4)
            self.__visualize_states(mat(i_w, i_h), labels=update_labels, means=update_means,
                                    covariance_matrices=update_covariance_matrices,
                                    existence_probabilities=update_existence_probabilities,
                                    visualize_covariance_matrix=False, visualize_existence_probabilities=True,
                                    visualize_state_size=True, original_image_size=original_image_size)
            self.__visualize_detections(mat(i_w, i_h), detections=all_detections, draw_detection_size=True,
                                        original_image_size=original_image_size)

        if False:
            i_w, i_h = 2, 0
            w, h = visualization_image.shape[:2:][::-1]
            vel_plot = self.__draw_state(self.tracks, w * 3, h)
            mat(i_w, i_h, vel_plot, s_h=3)

        self.__last_visualization = adapted_image

        return adapted_image

    def __draw_state(self, tracks: Dict[Tuple[int, int], Tuple[np.ndarray, float, Tuple[int, int, int]]], width,
                     height):

        plt.ioff()

        dpi = 90
        fig = plt.figure(figsize=(width / dpi, height / dpi), dpi=dpi)
        sp1 = fig.add_subplot(1, 3, 1)
        sp2 = fig.add_subplot(1, 3, 2)
        sp3 = fig.add_subplot(1, 3, 3)

        time_count = self.time_count + 1

        plt.subplots_adjust(left=0.012, bottom=0.01, right=1, top=0.9, wspace=0.08, hspace=0)

        sp1.scatter([time_count - 100, time_count], [0, 0])
        sp1.set_title('size')
        sp2.scatter([time_count - 100, time_count], [0, 0])
        sp2.set_title('orientation')
        sp3.scatter([time_count - 100, time_count], [0, 0])
        sp3.set_title('velocity')
        for lbl, track in tracks.items():
            means, _probability_of_existence, color = track

            means = means[-100:]

            velocity_x, velocity_y = means[:, -2:].T
            size_x, size_y = means[:, [2, 3]].T
            orientation = means[:, 4]

            color = [c / 255. for c in color]
            r = range(time_count - len(velocity_x), time_count)

            sp1.plot(r, size_x, linestyle='--', color=color, label=lbl)
            sp1.plot(r, size_y, color=color, label=lbl)

            sp2.plot(r, orientation, color=color)

            sp3.plot(r, velocity_x, linestyle='--', color=color)
            sp3.plot(r, velocity_y, color=color)

        sp2.set_ylim(-np.pi, np.pi)

        fig.legend()
        fig.canvas.draw()

        # Now we can save it to a numpy array.
        data = np.frombuffer(fig.canvas.tostring_rgb(), dtype=np.uint8)
        data = data.reshape(fig.canvas.get_width_height()[::-1] + (3,))
        fig.clf()
        plt.close()
        return data

# TODO: object death -> count
# - Strategy 1) if an object dies in the currently active hypothesis, remove it from all the other hypothesis and count.
#               This is what I will do for now
# - Strategy 2) If an object dies in the currently active hypothesis, check if it is inside a currently active
#               hypothesis. If it is

# TODO: Assignment matrix: set matches to infinity for combinations that are impossible
#       E.g. when a detection is matched behind another detection (if a pig would have had to walk through another pig).

# TODO:
# - visualization
# - Larger state space
# - Velocity model
#
