from typing import Tuple, Callable, Union

import numpy as np
import scipy.linalg


def compute_sigma_point_integral(means: np.ndarray, covariances: np.ndarray,
                                 g: Callable[[np.ndarray, Union[np.ndarray, float]], np.ndarray],
                                 return_poi: bool = False) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
    """
    Approximate \\E(g(x)) with x distributed according to \\N(\\cdot; means, covariances)

        => \\E(g(x)) =    \\int_x  g(x) dx
                     =    \\int_x  p_n(x; means, covariances) * g(x)
                 \\approx \\sum_i^n W_i g(x_i)

    The values of x_i and w_i are given by the sigma point integration method.
    We utilize the Cobature rule which computes as follows for a state space of dimensionality n.

        x_i = mean +- sqrt( n * cov[:,i] ) for i = 1 ... n
        W_i = 1 / 2n

    :param means:        The means / mean of the normal distribution(s) (shape (M x) d)
    :param covariances:  The covariance matrices / matrix of the normal distribution(s) (shape (M x) d x d)
    :param return_poi:   Set to true to receive  x_i as second return value (for visualization and debugging).
    :param g:            The function over which we would like to compute the expected value with regards to argument
                         x distributed according to the normal distributions parametrized via #means, #covariances.
    :return:             A numpy array that contains the approximation of \\E_N(g(x)) (shape M if return_poi is False)
    """

    # Extend covariances and means if only one covariance matrix is provided.
    if len(covariances.shape) == 2:
        covariances = covariances[np.newaxis]
        means = means[np.newaxis]

    number_of_gaussians: int = means.shape[0]
    dimensionality: int = means.shape[1]
    assert means.shape == (number_of_gaussians, dimensionality)
    assert covariances.shape == (number_of_gaussians, dimensionality, dimensionality)

    number_of_support_points: float = dimensionality * 2

    x_i: np.ndarray = np.repeat(means[None].astype(np.float64), dimensionality * 2, axis=0) \
        .reshape((number_of_support_points, number_of_gaussians, dimensionality))

    # x_i[i] == mu  \for all i
    assert ((x_i == means.astype(np.float64)[np.newaxis]).all())

    # xx[i] -+= sqrt(n * cov)[i]
    snc = np.swapaxes(np.array([scipy.linalg.sqrtm(dimensionality * covariances[n]) for n in range(number_of_gaussians)])\
        .reshape((number_of_gaussians, dimensionality, dimensionality)), 0, 1)

    #snc = np.swapaxes(scipy.linalg.sqrtm(dimensionality * covariances), 0, 1)
    if np.isnan(snc).any():
        raise 'nan'

    for i in range(number_of_gaussians):
        if np.abs(snc[:, i].dot(snc[:, i]) - dimensionality * covariances[i]).max(initial=0) >= 1e-9:
            raise 'not in range'
    assert not np.isnan(snc).any()
    assert np.abs(snc.dot(snc) - dimensionality * covariances).max(initial=0) < 1e-9
    assert snc.shape == (dimensionality, number_of_gaussians, dimensionality)
    x_i[:dimensionality] -= snc.T
    x_i[dimensionality:] += snc.T
    # x_i[i] == mu +- sqrt(n * cov[i])

    if __debug__:
        # Check the covariance matrix
        other_values = np.ones_like(x_i) * 42
        assert other_values.shape == (number_of_support_points, number_of_gaussians, dimensionality)
        for j in (-1, 1):
            for i in range(dimensionality):
                for i_g in range(number_of_gaussians):
                    index = i + (dimensionality if j == 1 else 0)
                    # covariances[i].shape == (dim x dim)
                    other_values[index, i_g] = \
                        means[i_g] + j * np.sqrt(dimensionality * covariances[i_g, :, i])

        diff = np.abs(other_values - x_i)
        assert (diff < 1e-10).all()

    # Evaluate the function at all the points
    g_x_i_n: np.ndarray = g(x_i, 1.0 / (2.0 * dimensionality))
    assert g_x_i_n.shape == (number_of_gaussians,)

    if return_poi:
        return g_x_i_n, x_i
    return g_x_i_n


def compute_sigma_point_integral_poi(means: np.ndarray, covariances: np.ndarray, roi, p_d=0.99,
                                     return_poi: bool = False) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
    """
    Approximate \\E(g(x)) with g(x) = \\1_{x \\in roi} * p_d with x distributed according
    to \\N(\\cdot; means, covariances)

        => \\E(g(x)) =    \\int_x  g(x) dx
                     =    \\int_x  p_n(x; means, covariances) * \\1_{x \\in roi} * p_d
                 \\approx \\sum_i^n W_i \\1_{x \\in roi} * p_d

    :param means:       @see compute_sigma_point_integral
    :param covariances: @see compute_sigma_point_integral
    :param return_poi:  @see compute_sigma_point_integral

    :param roi:   The region of interest, inside which the function g evaluates to `p_d`.
    :param p_d:   The value to which the function g evaluates inside the roi.

    :return:      A numpy array that contains the approximation of \\E_N(g(x)) (shape M if return_poi is False)
    """

    def g(x_i: np.ndarray, n_i: Union[float, np.ndarray]) -> np.ndarray:
        number_of_support_points: int = x_i.shape[0]
        number_of_gaussians: int = x_i.shape[1]
        state_dimensionality: int = x_i.shape[2]
        assert x_i.shape == (number_of_support_points, number_of_gaussians, state_dimensionality)
        assert type(n_i) is float or n_i.shape == (number_of_support_points,)

        area_ok = (roi[:2] <= x_i[:, :, :2]).all(axis=2) * (x_i[:, :, :2] <= roi[-2:]).all(axis=2)
        assert area_ok.shape == (number_of_support_points, number_of_gaussians)
        return (p_d * n_i) * np.sum(area_ok.T, axis=1)

    return compute_sigma_point_integral(means=means, covariances=covariances, g=g, return_poi=return_poi)


def compute_sigma_point_integral_map(means: np.ndarray, covariances: np.ndarray, g_map: np.ndarray, return_poi: bool = False) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
    """
    Approximate \\E(g(x)) with g(x) = \\1_{x \\in roi} * p_d with x distributed according
    to \\N(\\cdot; means, covariances)

        => \\E(g(x)) =    \\int_x  g(x) dx
                     =    \\int_x  p_n(x; means, covariances) * \\1_{x \\in roi} * p_d
                 \\approx \\sum_i^n W_i \\1_{x \\in roi} * p_d

    :param means:       @see compute_sigma_point_integral
    :param covariances: @see compute_sigma_point_integral
    :param return_poi:  @see compute_sigma_point_integral

    :param g_map:         Lookup map that contains the pmf of g

    :return:      A numpy array that contains the approximation of \\E_N(g(x)) (shape M if return_poi is False)
    """
    g_map_width: int = g_map.shape[0]
    g_map_height: int = g_map.shape[1]
    assert g_map.shape == (g_map_width, g_map_height)
    assert g_map.min(initial=0) >= 0 and g_map.max(initial=0) <= 1

    def g(x_i: np.ndarray, n_i: Union[float, np.ndarray]) -> np.ndarray:
        number_of_support_points: int = x_i.shape[0]
        number_of_gaussians: int = x_i.shape[1]
        state_dimensionality: int = x_i.shape[2]
        assert x_i.shape == (number_of_support_points, number_of_gaussians, state_dimensionality)
        assert isinstance(n_i, float) or isinstance(n_i, int) or n_i.shape == (number_of_support_points,)

        x_lookup: np.array = x_i.astype(np.int).reshape(number_of_support_points * number_of_gaussians,
                                                        state_dimensionality)
        x_lookup[:, 0] = x_lookup[:, 0].clip(min=0, max=g_map_width - 1)
        x_lookup[:, 1] = x_lookup[:, 1].clip(min=0, max=g_map_height - 1)

        looked_up: np.array = np.empty((number_of_support_points * number_of_gaussians), dtype=np.float64)
        for i in range(looked_up.shape[0]):
            looked_up[i] = g_map[x_lookup[i][0], x_lookup[i][1]]


        #looked_up: np.array = g_map[x_lookup]
        assert looked_up.shape == (number_of_support_points * number_of_gaussians, )

        g_evaluated: np.array = np.sum(looked_up.reshape(number_of_support_points, number_of_gaussians) * n_i, axis=0)
        assert g_evaluated.shape == (number_of_gaussians, )

        return g_evaluated

    # debug
    return g(means[:, :2][np.newaxis], 1), means[:, :2]

    #return compute_sigma_point_integral(means=means, covariances=covariances, g=g, return_poi=return_poi)


if __name__ == '__main__':

    import matplotlib.pyplot as plt


    def plot_heat_map(x, y, ww, hh, resolution, f, xx, probabilities, roi=None):
        plt.figure(figsize=(0.2 / resolution, 0.2 / resolution))
        grid = np.mgrid[x: x + ww: resolution, y: y + hh: resolution].T

        w, h = grid.shape[0], grid.shape[1]
        grid = grid.reshape([w * h, grid.shape[2]]).astype('float64')
        v, s = f(grid)
        values = v.reshape((w, h))

        plt.imshow(values, extent=(x, ww + x, y + hh, y))

        xxx, yyy = xx.T

        correct = []

        print(values.shape)

        for i, (xs, ys) in enumerate(zip(xxx, yyy)):
            plt.scatter(xs, ys, label=f'{i}, {probabilities[i]} {s[i] * resolution ** 2}')

        if roi is not None:
            plt.plot([roi[0], roi[2], roi[2], roi[0], roi[0]], [roi[1], roi[1], roi[3], roi[3], roi[1]], color='black',
                     linewidth=5)

        plt.legend()

        plt.show()


    def visualize_sigma_point_integrals():

        cov = np.array([[0.5, 0], [0, 1]])
        means = np.array([[1, 1], [3, 3], [7, 8]]) / 1 + [1, 1]
        covs = np.vstack((cov[None],) * len(means))

        covs = np.array([[[0.9, .1], [.5, 2]], [[2.0, .1], [.1, 0.2]], [[1.1, 1.0], [0.9, 1.1]]]) / 2

        roi = [0, 0, 10.0, 10.0]

        probabilities, points = compute_sigma_point_integral(means=means, covs=covs, roi=roi)

        def compute(x: np.ndarray):
            sus = []
            su = None
            invs = np.linalg.inv(covs)
            divisors = np.sqrt(2 * np.pi ** 2 * np.linalg.det(covs))
            for i, m in enumerate(means):
                diff = x - means[i]
                id = np.sum(np.dot(invs[i], diff.T).T * diff, axis=1)
                val = 1 * np.exp(-1 / 2. * id) / divisors[i]
                if su is None:
                    su = val
                else:
                    su += val
                sus += [np.sum(val)]

            return su, sus

        plot_heat_map(-1, -1, roi[2] + 2, roi[3] + 2, 0.05, compute, xx=points, probabilities=probabilities, roi=roi)


    visualize_sigma_point_integrals()
