from albumentations import Crop
from typing import Optional


class RelativeCrop(Crop):

    def __init__(self, x_min_percent: float = 0.0, y_min_percent: float = 0.0,
                 x_max_percent: float = 1.0, y_max_percent=1.0, always_apply=False, p=1.0):
        super().__init__(x_min=0, y_min=0, x_max=1, y_max=1, always_apply=always_apply, p=p)

        # In development mode ensure compatibility with the implementation of crop shipped with the version of
        # Albumentation that is currently in-use.
        if any((getattr(self, var, None) is None for var in ('x_min', 'x_max', 'y_min', 'y_max'))):
            raise NotImplemented('This version of Albumentations is incompatible with the RelativeCrop augmentation.'
                                 'Make sure that you use the recommended version of Albumentations and report'
                                 'this incompatibility to a DataSpree developer.')

        self.x_min_percent: float = x_min_percent
        self.y_min_percent: float = y_min_percent
        self.x_max_percent: float = x_max_percent
        self.y_max_percent: float = y_max_percent

        if not (0 <= self.x_min_percent <= self.x_max_percent <= 1):
            raise ValueError(f'The x coordinates for the Transformation "{type(self).__name__}" are invalid:'
                             f'"{0} <= {self.x_min_percent} = x min [%] <= {self.x_max_percent} = x max [%] <= 1" '
                             f'does not hold')

        if not (0 <= self.y_min_percent <= self.y_max_percent <= 1):
            raise ValueError(f'The y coordinates for the Transformation "{type(self).__name__}" are invalid:'
                             f'"{0} <= {self.y_min_percent} = y min [%] <= {self.y_max_percent} = y max [%] <= 1" '
                             f'does not hold')
        self.__last_input_width: int = 0
        self.__last_input_height: int = 0

    def update_params(self, params, **kwargs):
        # @dev I think update_params is the right choice here, if not check out & implement
        # def get_params_dependent_on_targets(self, params: Dict[str, Any]) -> Dict[str, Any]:

        params = super(RelativeCrop, self).update_params(params, **kwargs)
        self.__last_input_height = params['rows']
        self.__last_input_width = params['cols']

        self.x_min = round(self.x_min_percent * self.__last_input_width)
        self.y_min = round(self.y_min_percent * self.__last_input_height)
        self.x_max = max(round(self.x_max_percent * self.__last_input_width), self.x_min + 1)
        self.y_max = max(round(self.y_max_percent * self.__last_input_height), self.y_min + 1)

        # XXX: kriegen wir das hier noch schöner hin?
        if self.y_max > self.__last_input_height:
            self.y_max -= 1
            self.y_min -= 1

        if self.x_max > self.__last_input_width:
            self.x_max -= 1
            self.x_min -= 1

        return super(RelativeCrop, self).update_params(params, **kwargs)

    def __check_after_update(self, rows: Optional[int] = None, cols: Optional[int] = None, **_kwargs):
        if rows is None and cols is None:
            raise NotImplemented('Rows and Columns have to be provided by a compatible version of Albumentations. '
                                 'Make sure you use the recommended version of Albumentations and report this '
                                 'incompatibility to a DataSpree developer.')
        else:
            if rows != self.__last_input_height:
                raise NotImplemented(f'Rows {rows} does not match expected height {self.__last_input_height}. '
                                     f'The current version of Albumentations is incompatible with the '
                                     f'{type(self).__name__} augmentation.'
                                     'Make sure that you use the recommended version of Albumentations and report '
                                     'this incompatibility to a DataSpree developer.')
            if cols != self.__last_input_width:
                raise NotImplemented(f'Rows {cols} does not match expected height {self.__last_input_width}'
                                     f'The current version of Albumentations is incompatible with the '
                                     f'{type(self).__name__} augmentation.'
                                     'Make sure that you use the recommended version of Albumentations and report '
                                     'this incompatibility to a DataSpree developer.')

    def apply(self, img, **params):
        self.__check_after_update(**params)
        return super(RelativeCrop, self).apply(img, **params)

    def apply_to_bbox(self, bbox, **params):
        self.__check_after_update(**params)
        return super(RelativeCrop, self).apply_to_bbox(bbox, **params)

    def apply_to_keypoint(self, keypoint, **params):
        self.__check_after_update(**params)
        return super(RelativeCrop, self).apply_to_keypoint(keypoint, **params)

    def get_transform_init_args_names(self):
        # @dev: One could also argue not overriding this, and allowing the relative crop to optionally receive
        #       x_min, y_min, x_max, y_max (I think this is relevant for replay, and replaying an augmentation
        #       should probably require the input shape not to change.
        return 'x_min_percent', 'y_min_percent', 'x_max_percent', 'y_max_percent'

    @classmethod
    def get_class_fullname(cls):
        return cls.__name__
